"""
Helper methods to unzip an archive preserving the file permissions. Python's zipfile module does not yet support
this feature natively (https://bugs.python.org/issue15795).
"""

import logging
import os
import zipfile
from pathlib import Path

import requests

from samcli.lib.utils.progressbar import progressbar

LOG = logging.getLogger(__name__)

S_IFLNK = 0xA


def _is_symlink(file_info):
    """
    Check the upper 4 bits of the external attribute for a symlink.
    See: https://unix.stackexchange.com/questions/14705/the-zip-formats-external-file-attribute

    Parameters
    ----------
    file_info : zipfile.ZipInfo
        The ZipInfo for a ZipFile

    Returns
    -------
    bool
        A response regarding whether the ZipInfo defines a symlink or not.
    """

    return (file_info.external_attr >> 28) == 0xA  # noqa: PLR2004


def _extract(file_info, output_dir, zip_ref):
    """
    Unzip the given file into the given directory while preserving file permissions in the process.

    Parameters
    ----------
    file_info : zipfile.ZipInfo
        The ZipInfo for a ZipFile
    output_dir : str
        Path to the directory where the it should be unzipped to
    zip_ref : zipfile.ZipFile
        The ZipFile we are working with.

    Returns
    -------
    string
        Returns the target path the Zip Entry was extracted to.
    """

    # Handle any regular file/directory entries
    if not _is_symlink(file_info):
        return zip_ref.extract(file_info, output_dir)

    source = zip_ref.read(file_info.filename).decode("utf8")
    link_name = os.path.normpath(os.path.join(output_dir, file_info.filename))

    # make leading dirs if needed
    leading_dirs = os.path.dirname(link_name)
    if not os.path.exists(leading_dirs):
        os.makedirs(leading_dirs)

    # If the link already exists, delete it or symlink() fails
    if os.path.lexists(link_name):
        os.remove(link_name)

    # Create a symbolic link pointing to source named link_name.
    os.symlink(source, link_name)

    return link_name


def unzip(zip_file_path, output_dir, permission=None):
    """
    Unzip the given file into the given directory while preserving file permissions in the process.

    Parameters
    ----------
    zip_file_path : str
        Path to the zip file
    output_dir : str
        Path to the directory where the it should be unzipped to
    permission : int
        Permission to set in an octal int form
    """

    with zipfile.ZipFile(zip_file_path, "r") as zip_ref:

        # For each item in the zip file, extract the file and set permissions if available
        for file_info in zip_ref.infolist():
            extracted_path = _extract(file_info, output_dir, zip_ref)

            # If the extracted_path is a symlink, do not set the permissions. If the target of the symlink does not
            # exist, then os.chmod will fail with FileNotFoundError
            if not os.path.islink(extracted_path):
                _set_permissions(file_info, extracted_path)
                _override_permissions(extracted_path, permission)

    if not os.path.islink(extracted_path):
        _override_permissions(output_dir, permission)


def _override_permissions(path, permission):
    """
    Forcefully override the permissions on the path

    Parameters
    ----------
    path str
        Path where the file or directory
    permission octal int
        Permission to set

    """
    if permission:
        os.chmod(path, permission)


def _set_permissions(zip_file_info, extracted_path):
    """
    Sets permissions on the extracted file by reading the ``external_attr`` property of given file info.

    Parameters
    ----------
    zip_file_info : zipfile.ZipInfo
        Object containing information about a file within a zip archive

    extracted_path : str
        Path where the file has been extracted to
    """

    # Permission information is stored in first two bytes.
    permission = zip_file_info.external_attr >> 16
    if not permission:
        # Zips created on certain Windows machines, however, might not have any permission information on them.
        # Skip setting a permission on these files.
        LOG.debug("File %s in zipfile does not have permission information", zip_file_info.filename)
        return

    os.chmod(extracted_path, permission)


def unzip_from_uri(uri, layer_zip_path, unzip_output_dir, progressbar_label):
    """
    Download the LayerVersion Zip to the Layer Pkg Cache

    Parameters
    ----------
    uri str
        Uri to download from
    layer_zip_path str
        Path to where the content from the uri should be downloaded to
    unzip_output_dir str
        Path to unzip the zip to
    progressbar_label str
        Label to use in the Progressbar
    """
    try:
        get_request = requests.get(uri, stream=True, verify=os.environ.get("AWS_CA_BUNDLE", True))

        with open(layer_zip_path, "wb") as local_layer_file:
            file_length = int(get_request.headers["Content-length"])

            with progressbar(file_length, progressbar_label) as p_bar:
                # Set the chunk size to None. Since we are streaming the request, None will allow the data to be
                # read as it arrives in whatever size the chunks are received.
                for data in get_request.iter_content(chunk_size=None):
                    local_layer_file.write(data)
                    p_bar.update(len(data))

        # Forcefully set the permissions to 700 on files and directories. This is to ensure the owner
        # of the files is the only one that can read, write, or execute the files.
        unzip(layer_zip_path, unzip_output_dir, permission=0o700)

    finally:
        # Remove the downloaded zip file
        path_to_layer = Path(layer_zip_path)
        if path_to_layer.exists():
            path_to_layer.unlink()



"""
Stream cursor utilities for moving cursor in the terminal.
"""
import os
import platform

# NOTE: ANSI escape codes.
# NOTE: Still needs investigation on non terminal environments.
ESC = "\u001B["

# Enables ANSI escape codes on Windows
if platform.system().lower() == "windows":
    try:
        os.system("color")
    except Exception:
        pass


class CursorFormatter:
    """
    Base class for defining how cursor is to be manipulated.
    """

    def __init__(self):
        pass

    def cursor_format(self, count):
        pass


class CursorUpFormatter(CursorFormatter):
    """
    Class for formatting and outputting moving the cursor up within the stream of bytes.
    """

    def cursor_format(self, count=0):
        return ESC + str(count) + "A"


class CursorDownFormatter(CursorFormatter):
    """
    Class for formatting and outputting moving the cursor down within the stream of bytes.
    """

    def cursor_format(self, count=0):
        return ESC + str(count) + "B"


class ClearLineFormatter(CursorFormatter):
    """
    Class for formatting and outputting clearing the cursor within the stream of bytes.
    """

    def cursor_format(self, count=0):
        return ESC + str(count) + "K"


class CursorLeftFormatter(CursorFormatter):
    """
    Class for formatting and outputting moving the cursor left within the stream of bytes.
    """

    def cursor_format(self, count=0):
        return ESC + str(count) + "G"



"""
Class to store the API configurations in the SAM Template. This class helps store both implicit and explicit
routes in a standardized format
"""

import logging
from collections import defaultdict
from typing import Dict, Iterator, List, Optional, Set, Tuple, Union

from samcli.lib.providers.provider import Api, Cors
from samcli.local.apigw.local_apigw_service import Route

LOG = logging.getLogger(__name__)


class ApiCollector:
    def __init__(self) -> None:
        # Route properties stored per resource.
        self._route_per_resource: Dict[str, List[Route]] = defaultdict(list)

        # processed values to be set before creating the api
        self._routes: List[Route] = []
        self.binary_media_types_set: Set[str] = set()
        self.stage_name: Optional[str] = None
        self.stage_variables: Optional[Dict] = None
        self.cors: Optional[Cors] = None

    def __iter__(self) -> Iterator[Tuple[str, List[Route]]]:
        """
        Iterator to iterate through all the routes stored in the collector. In each iteration, this yields the
        LogicalId of the route resource and a list of routes available in this resource.
        Yields
        -------
        str
            LogicalID of the AWS::Serverless::Api or AWS::ApiGateway::RestApi resource
        list samcli.commands.local.lib.provider.Api
            List of the API available in this resource along with additional configuration like binary media types.
        """

        for logical_id, _ in self._route_per_resource.items():
            yield logical_id, self._get_routes(logical_id)

    def add_routes(self, logical_id: str, routes: List[Route]) -> None:
        """
        Stores the given routes tagged under the given logicalId
        Parameters
        ----------
        logical_id : str
            LogicalId of the AWS::Serverless::Api or AWS::ApiGateway::RestApi resource
        routes : list of samcli.commands.local.agiw.local_apigw_service.Route
            List of routes available in this resource
        """
        self._get_routes(logical_id).extend(routes)

    def _get_routes(self, logical_id: str) -> List[Route]:
        """
        Returns the properties of resource with given logical ID. If a resource is not found, then it returns an
        empty data.
        Parameters
        ----------
        logical_id : str
            Logical ID of the resource
        Returns
        -------
        samcli.commands.local.lib.Routes
            Properties object for this resource.
        """

        return self._route_per_resource[logical_id]

    @property
    def routes(self) -> List[Route]:
        return self._routes if self._routes else self.all_routes()

    @routes.setter
    def routes(self, routes: List[Route]) -> None:
        self._routes = routes

    def all_routes(self) -> List[Route]:
        """
        Gets all the routes within the _route_per_resource

        Return
        -------
        All the routes within the _route_per_resource
        """
        routes = []
        for logical_id in self._route_per_resource.keys():
            routes.extend(self._get_routes(logical_id))
        return routes

    def get_api(self) -> Api:
        """
        Creates the api using the parts from the ApiCollector. The routes are also deduped so that there is no
        duplicate routes with the same function name, path, but different method.

        The normalised_routes are the routes that have been processed. By default, this will get all the routes.
        However, it can be changed to override the default value of normalised routes such as in SamApiProvider

        Return
        -------
        An Api object with all the properties
        """
        api = Api()
        routes = self.dedupe_function_routes(self.routes)
        routes = self.normalize_cors_methods(routes, self.cors)
        api.routes = routes
        api.binary_media_types_set = self.binary_media_types_set
        api.stage_name = self.stage_name
        api.stage_variables = self.stage_variables
        api.cors = self.cors
        return api

    @staticmethod
    def normalize_cors_methods(routes: List[Route], cors: Optional[Cors]) -> List[Route]:
        """
        Adds OPTIONS method to all the route methods if cors exists

        Parameters
        -----------
        routes: list(samcli.local.apigw.local_apigw_service.Route)
            List of Routes

        cors: samcli.commands.local.lib.provider.Cors
            the cors object for the api

        Return
        -------
        A list of routes without duplicate routes with the same function_name and method
        """

        def add_options_to_route(route: Route) -> Route:
            if "OPTIONS" not in route.methods:
                route.methods.append("OPTIONS")
            return route

        return routes if not cors else [add_options_to_route(route) for route in routes]

    @staticmethod
    def dedupe_function_routes(routes: List[Route]) -> List[Route]:
        """
         Remove duplicate routes that have the same function_name and method

         route: list(Route)
             List of Routes

        Return
        -------
        A list of routes without duplicate routes with the same stack_path, function_name and method
        """
        grouped_routes: Dict[str, Route] = {}

        for route in routes:
            key = "{}-{}-{}-{}".format(route.stack_path, route.function_name, route.path, route.operation_name or "")
            config = grouped_routes.get(key, None)
            methods = route.methods
            if config:
                methods += config.methods
            sorted_methods = sorted(methods)
            grouped_routes[key] = Route(
                function_name=route.function_name,
                path=route.path,
                methods=sorted_methods,
                event_type=route.event_type,
                payload_format_version=route.payload_format_version,
                operation_name=route.operation_name,
                stack_path=route.stack_path,
            )
        return list(grouped_routes.values())

    def add_binary_media_types(self, logical_id: str, binary_media_types: Optional[List[str]]) -> None:
        """
        Stores the binary media type configuration for the API with given logical ID
        Parameters
        ----------

        logical_id : str
            LogicalId of the AWS::Serverless::Api resource

        binary_media_types : list of str
            List of binary media types supported by this resource
        """

        binary_media_types = binary_media_types or []
        for value in binary_media_types:
            normalized_value = self.normalize_binary_media_type(value)

            # If the value is not supported, then just skip it.
            if normalized_value:
                self.binary_media_types_set.add(normalized_value)
            else:
                LOG.debug("Unsupported data type of binary media type value of resource '%s'", logical_id)

    @staticmethod
    def normalize_binary_media_type(value: Union[str, Dict]) -> Optional[str]:
        """
        Converts binary media types values to the canonical format. Ex: image~1gif -> image/gif. If the value is not
        a string, then this method just returns None
        Parameters
        ----------
        value
            Value to be normalized. Expect to be a string.
            However, it is possible that user specified a non-str (dict) value for one of the binary media types.
            If so, return None.
        Returns
        -------
        str or None
            Normalized value. If the input was not a string, then None is returned
        """

        if not isinstance(value, str):
            return None

        return value.replace("~1", "/")



"""
    Tests cookiecutter baking process and rendered content
"""


def test_project_tree(cookies):
    result = cookies.bake(extra_context={"project_name": "hello sam"})
    assert result.exit_code == 0
    assert result.exception is None
    assert result.project.basename == "hello sam"
    assert result.project.isdir()
    assert result.project.join(".gitignore").isfile()
    assert result.project.join("template.yaml").isfile()
    assert result.project.join("README.md").isfile()
    assert result.project.join("src").isdir()
    assert result.project.join("test").isdir()
    assert result.project.join("src", "HelloWorld").isdir()
    assert result.project.join("src", "HelloWorld", "HelloWorld.csproj").isfile()
    assert result.project.join("src", "HelloWorld", "Function.cs").isfile()
    assert result.project.join("src", "HelloWorld", "aws-lambda-tools-defaults.json").isfile()
    assert result.project.join("test", "HelloWorld.Test", "FunctionTest.cs").isfile()
    assert result.project.join("test", "HelloWorld.Test", "HelloWorld.Tests.csproj").isfile()


def test_app_content(cookies):
    result = cookies.bake(extra_context={"project_name": "my_lambda"})
    app_file = result.project.join("src", "HelloWorld", "Function.cs")
    app_content = app_file.readlines()
    app_content = "".join(app_content)

    contents = ("GetCallingIP", "GetStringAsync", "location", "message", "hello world", "StatusCode")

    for content in contents:
        assert content in app_content



"""
Hash calculation utilities for files and directories.
"""
import hashlib
import os
from typing import Any, List, Optional, cast

BLOCK_SIZE = 4096


def file_checksum(file_name: str, hash_generator: Any = None) -> str:
    """

    Parameters
    ----------
    file_name : str
        file name of the file for which md5 checksum is required.
    hash_generator : hashlib._Hash
        hashlib _Hash object for generating hashes. Defaults to hashlib.md5.

    Returns
    -------
    checksum of the given file.

    """
    # Default value is set here because default values are static mutable in Python
    if not hash_generator:
        hash_generator = hashlib.md5()
    with open(file_name, "rb") as file_handle:
        # Save current cursor position and reset cursor to start of file
        curpos = file_handle.tell()
        file_handle.seek(0)

        buf = file_handle.read(BLOCK_SIZE)
        while buf:
            hash_generator.update(buf)
            buf = file_handle.read(BLOCK_SIZE)

        # Restore file cursor's position
        file_handle.seek(curpos)

        return cast(str, hash_generator.hexdigest())


def dir_checksum(
    directory: str, followlinks: bool = True, ignore_list: Optional[List[str]] = None, hash_generator: Any = None
) -> str:
    """

    Parameters
    ----------
    directory : dict
        A directory with an absolute path
    followlinks : bool
        Follow symbolic links through the given directory
    ignore_list : list(str)
        The list of file/directory names to ignore in checksum
    hash_generator : hashlib._Hash
        The hashing method (hashlib _Hash object) that generates checksum. Defaults to hashlib.md5.

    Returns
    -------
    checksum hash of the directory.

    """
    ignore_set = set(ignore_list or [])
    if not hash_generator:
        hash_generator = hashlib.md5()
    files = list()
    # Walk through given directory and find all directories and files.
    for dirpath, dirnames, filenames in os.walk(directory, followlinks=followlinks):
        # > When topdown is True, the caller can modify the dirnames list in-place
        # > (perhaps using del or slice assignment) and walk() will only recurse
        # > into the subdirectories whose names remain in dirnames
        # > https://docs.python.org/library/os.html#os.walk
        dirnames[:] = [dirname for dirname in dirnames if dirname not in ignore_set]
        # Go through every file in the directory and sub-directory.
        for filepath in [os.path.join(dirpath, filename) for filename in filenames if filename not in ignore_set]:
            # Look at filename and contents.
            # Encode file's checksum to be utf-8 and bytes.
            files.append(filepath)

    files.sort()
    for file in files:
        hash_generator.update(os.path.relpath(file, directory).encode("utf-8"))
        filepath_checksum = file_checksum(file)
        hash_generator.update(filepath_checksum.encode("utf-8"))

    return cast(str, hash_generator.hexdigest())


def str_checksum(content: str) -> str:
    """
    return a md5 checksum of a given string

    Parameters
    ----------
    content: string
        the string to be hashed
    Returns
    -------
    md5 checksum of content
    """
    md5 = hashlib.md5()
    md5.update(content.encode("utf-8"))
    return md5.hexdigest()



"""
Class representing the samconfig.toml
"""

import logging
import os
from pathlib import Path
from typing import Any, Iterable

import tomlkit

from samcli.lib.config.exceptions import SamConfigVersionException
from samcli.lib.config.version import SAM_CONFIG_VERSION, VERSION_KEY

LOG = logging.getLogger(__name__)

DEFAULT_CONFIG_FILE_EXTENSION = "toml"
DEFAULT_CONFIG_FILE_NAME = f"samconfig.{DEFAULT_CONFIG_FILE_EXTENSION}"
DEFAULT_ENV = "default"
DEFAULT_GLOBAL_CMDNAME = "global"


class SamConfig:
    """
    Class to interface with `samconfig.toml` file.
    """

    document = None

    def __init__(self, config_dir, filename=None):
        """
        Initialize the class

        Parameters
        ----------
        config_dir : string
            Directory where the configuration file needs to be stored
        filename : string
            Optional. Name of the configuration file. It is recommended to stick with default so in the future we
            could automatically support auto-resolving multiple config files within same directory.
        """
        self.filepath = Path(config_dir, filename or DEFAULT_CONFIG_FILE_NAME)

    def get_stage_configuration_names(self):
        self._read()
        if isinstance(self.document, dict):
            return [stage for stage, value in self.document.items() if isinstance(value, dict)]
        return []

    def get_all(self, cmd_names, section, env=DEFAULT_ENV):
        """
        Gets a value from the configuration file for the given environment, command and section

        Parameters
        ----------
        cmd_names : list(str)
            List of representing the entire command. Ex: ["local", "generate-event", "s3", "put"]
        section : str
            Specific section within the command to look into. e.g. `parameters`
        env : str
            Optional, Name of the environment

        Returns
        -------
        dict
            Dictionary of configuration options in the file. None, if the config doesn't exist.

        Raises
        ------
        KeyError
            If the config file does *not* have the specific section

        tomlkit.exceptions.TOMLKitError
            If the configuration file is invalid
        """

        env = env or DEFAULT_ENV

        self._read()
        if isinstance(self.document, dict):
            toml_content = self.document.get(env, {})
            params = toml_content.get(self._to_key(cmd_names), {}).get(section, {})
            if DEFAULT_GLOBAL_CMDNAME in toml_content:
                global_params = toml_content.get(DEFAULT_GLOBAL_CMDNAME, {}).get(section, {})
                global_params.update(params.copy())
                params = global_params.copy()
            return params
        return {}

    def put(self, cmd_names, section, key, value, env=DEFAULT_ENV):
        """
        Writes the `key=value` under the given section. You have to call the `flush()` method after `put()` in
        order to write the values back to the config file. Otherwise they will be just saved in-memory, available
        for future access, but never saved back to the file.

        Parameters
        ----------
        cmd_names : list(str)
            List of representing the entire command. Ex: ["local", "generate-event", "s3", "put"]
        section : str
            Specific section within the command to look into. e.g. `parameters`
        key : str
            Key to write the data under
        value : Any
            Value to write. Could be any of the supported TOML types.
        env : str
            Optional, Name of the environment

        Raises
        ------
        tomlkit.exceptions.TOMLKitError
            If the data is invalid
        """

        if self.document is None:
            # Checking for None here since a TOMLDocument can include a
            # 'body' property but still be falsy without a 'value' property
            self._read()
        # Empty document prepare the initial structure.
        # self.document is a nested dict, we need to check each layer and add new tables, otherwise duplicated key
        # in parent layer will override the whole child layer
        cmd_name_key = self._to_key(cmd_names)
        env_content = self.document.get(env, {})
        cmd_content = env_content.get(cmd_name_key, {})
        param_content = cmd_content.get(section, {})
        if param_content:
            param_content.update({key: value})
        elif cmd_content:
            cmd_content.update({section: {key: value}})
        elif env_content:
            env_content.update({cmd_name_key: {section: {key: value}}})
        else:
            self.document.update({env: {cmd_name_key: {section: {key: value}}}})
        # If the value we want to add to samconfig already exist in global section, we don't put it again in
        # the special command section
        self._deduplicate_global_parameters(cmd_name_key, section, key, env)

    def put_comment(self, comment):
        """
        Write a comment section back to the file.

        Parameters
        ------
        comment: str
            A comment to write to the samconfg file
        """
        if self.document is None:
            self._read()

        self.document.add(tomlkit.comment(comment))

    def flush(self):
        """
        Write the data back to file

        Raises
        ------
        tomlkit.exceptions.TOMLKitError
            If the data is invalid

        """
        self._write()

    def sanity_check(self):
        """
        Sanity check the contents of samconfig
        """
        try:
            self._read()
        except tomlkit.exceptions.TOMLKitError:
            return False
        else:
            return True

    def exists(self):
        return self.filepath.exists()

    def _ensure_exists(self):
        self.filepath.parent.mkdir(parents=True, exist_ok=True)
        self.filepath.touch()

    def path(self):
        return str(self.filepath)

    @staticmethod
    def config_dir(template_file_path=None):
        """
        SAM Config file is always relative to the SAM Template. If it the template is not
        given, then it is relative to cwd()
        """
        if template_file_path:
            return os.path.dirname(template_file_path)

        return os.getcwd()

    def _read(self):
        if not self.document:
            try:
                txt = self.filepath.read_text()
                self.document = tomlkit.loads(txt)
                self._version_sanity_check(self._version())
            except OSError:
                self.document = tomlkit.document()

        if self.document.body:
            self._version_sanity_check(self._version())
        return self.document

    def _write(self):
        if not self.document:
            return

        self._ensure_exists()

        current_version = self._version() if self._version() else SAM_CONFIG_VERSION
        try:
            self.document.add(VERSION_KEY, current_version)
        except tomlkit.exceptions.KeyAlreadyPresent:
            # NOTE(TheSriram): Do not attempt to re-write an existing version
            pass
        self.filepath.write_text(tomlkit.dumps(self.document))

    def _version(self):
        return self.document.get(VERSION_KEY, None)

    def _deduplicate_global_parameters(self, cmd_name_key, section, key, env=DEFAULT_ENV):
        """
        In case the global parameters contains the same key-value pair with command parameters,
        we only keep the entry in global parameters

        Parameters
        ----------
        cmd_name_key : str
            key of command name

        section : str
            Specific section within the command to look into. e.g. `parameters`

        key : str
            Key to write the data under

        env : str
            Optional, Name of the environment
        """
        global_params = self.document.get(env, {}).get(DEFAULT_GLOBAL_CMDNAME, {}).get(section, {})
        command_params = self.document.get(env, {}).get(cmd_name_key, {}).get(section, {})
        if (
            cmd_name_key != DEFAULT_GLOBAL_CMDNAME
            and global_params
            and command_params
            and global_params.get(key)
            and global_params.get(key) == command_params.get(key)
        ):
            value = command_params.get(key)
            save_global_message = (
                f'\n\tParameter "{key}={value}" in [{env}.{cmd_name_key}.{section}] is defined as a global '
                f"parameter [{env}.{DEFAULT_GLOBAL_CMDNAME}.{section}].\n\tThis parameter will be only saved "
                f"under [{env}.{DEFAULT_GLOBAL_CMDNAME}.{section}] in {self.filepath}."
            )
            LOG.info(save_global_message)
            # Only keep the global parameter
            del self.document[env][cmd_name_key][section][key]

    @staticmethod
    def _version_sanity_check(version: Any) -> None:
        if not isinstance(version, float):
            raise SamConfigVersionException(f"'{VERSION_KEY}' key is not present or is in unrecognized format. ")

    @staticmethod
    def _to_key(cmd_names: Iterable[str]) -> str:
        # construct a parsed name that is of the format: a_b_c_d
        return "_".join([cmd.replace("-", "_").replace(" ", "_") for cmd in cmd_names])



"""
Log streaming utilities when streaming logs from Docker
"""
import os
from typing import Dict

import docker

from samcli.lib.package.stream_cursor_utils import (
    ClearLineFormatter,
    CursorDownFormatter,
    CursorLeftFormatter,
    CursorUpFormatter,
)
from samcli.lib.utils.stream_writer import StreamWriter


class LogStreamError(Exception):
    def __init__(self, msg: str) -> None:
        Exception.__init__(self, msg)


class LogStreamer:
    def __init__(self, stream: StreamWriter):
        self._stream = stream
        self._cursor_up_formatter = CursorUpFormatter()
        self._cursor_down_formatter = CursorDownFormatter()
        self._cursor_left_formatter = CursorLeftFormatter()
        self._cursor_clear_formatter = ClearLineFormatter()

    def stream_progress(self, logs: docker.APIClient.logs):
        """
        Stream progress from docker push logs and move the cursor based on the log id.
        :param logs: generator from docker_clent.APIClient.logs
        """
        ids: Dict[str, int] = dict()
        for log in logs:
            _id = log.get("id", "")
            status = log.get("status", "")
            stream = log.get("stream", "")
            progress = log.get("progress", "")
            error = log.get("error", "")
            change_cursor_count = 0
            if _id:
                if _id not in ids:
                    ids[_id] = len(ids)
                else:
                    curr_log_line_id = ids[_id]
                    change_cursor_count = len(ids) - curr_log_line_id
                    self._stream.write(
                        self._cursor_up_formatter.cursor_format(change_cursor_count)
                        + self._cursor_left_formatter.cursor_format(),
                        encode=True,
                    )

            self._stream_write(_id, status, stream, progress, error)

            if _id:
                self._stream.write(
                    self._cursor_down_formatter.cursor_format(change_cursor_count)
                    + self._cursor_left_formatter.cursor_format(),
                    encode=True,
                )
        self._stream.write(os.linesep, encode=True)

    def _stream_write(self, _id: str, status: str, stream: bytes, progress: str, error: str):
        """
        Write stream information to stderr, if the stream information contains a log id,
        use the carriage return character to rewrite that particular line.
        :param _id: docker log id
        :param status: docker log status
        :param stream: stream, usually stderr
        :param progress: docker log progress
        :param error: docker log error
        """
        if error:
            raise LogStreamError(msg=error)
        if not status and not stream:
            return

        # NOTE(sriram-mv): Required for the purposes of when the cursor overflows existing terminal buffer.
        if not stream:
            self._stream.write(os.linesep, encode=True)
            self._stream.write(
                self._cursor_up_formatter.cursor_format() + self._cursor_left_formatter.cursor_format(), encode=True
            )
            self._stream.write(self._cursor_clear_formatter.cursor_format(), encode=True)

        if not _id:
            self._stream.write(stream, encode=True)
            self._stream.write(status, encode=True)
        else:
            self._stream.write(f"\r{_id}: {status} {progress}", encode=True)



"""
Provide a CFN implementation of IaCPluginInterface
"""

import logging
import os
from typing import List, Optional
from urllib.parse import unquote, urlparse

import jmespath

from samcli.commands._utils.template import get_template_data
from samcli.lib.iac.constants import GLOBAL_PARAMETER_OVERRIDES, PARAMETER_OVERRIDES
from samcli.lib.iac.plugins_interfaces import (
    Asset,
    DictSection,
    DictSectionItem,
    IaCPluginInterface,
    ImageAsset,
    LookupPath,
    S3Asset,
    SamCliContext,
    SamCliProject,
    Stack,
)
from samcli.lib.providers.sam_base_provider import SamBaseProvider
from samcli.lib.providers.sam_stack_provider import SamLocalStackProvider, get_local_path, is_local_path
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.resources import (
    AWS_LAMBDA_FUNCTION,
    AWS_LAMBDA_LAYERVERSION,
    AWS_SERVERLESS_FUNCTION,
    AWS_SERVERLESS_LAYERVERSION,
    METADATA_WITH_LOCAL_PATHS,
    NESTED_STACKS_RESOURCES,
    RESOURCES_WITH_IMAGE_COMPONENT,
    RESOURCES_WITH_LOCAL_PATHS,
)

LOG = logging.getLogger(__name__)

PARENT_STACK_TEMPLATE_PATH_KEY = "parent_stack_template_path"
TEMPLATE_PATH_KEY = "template_path"
TEMPLATE_BUILD_PATH_KEY = "template_build_path"

BASE_DIR_RESOURCES = [
    AWS_SERVERLESS_FUNCTION,
    AWS_LAMBDA_FUNCTION,
    AWS_SERVERLESS_LAYERVERSION,
    AWS_LAMBDA_LAYERVERSION,
]


class CfnIacImplementation(IaCPluginInterface):
    """
    CFN implementation for the plugins interface.
    read_project parses the CFN and returns a SamCliProject object
    write_project writes the updated project
        back to the build dir and returns true if successful
    update_packaged_locations updates the package locations and
        returns true if successful
    get_iac_file_types returns a list of file types/patterns associated with
        the CFN project type
    """

    def __init__(self, context: SamCliContext):
        self._template_file = context.command_options_map["template_file"]
        self._base_dir = context.command_options_map.get("base_dir", None)
        super().__init__(context)

    def read_project(self, lookup_paths: List[LookupPath]) -> SamCliProject:
        stack = self._build_stack(self._template_file)
        return SamCliProject([stack])

    def write_project(self, project: SamCliProject, build_dir: str) -> bool:
        # TODO
        pass

    def update_packaged_locations(self, stack: Stack) -> bool:
        # TODO
        pass

    # pylint: disable=too-many-branches
    def _build_stack(self, path: str, is_nested: bool = False, name: Optional[str] = None) -> Stack:
        asset: Asset
        assets: List[Asset] = []

        if os.path.islink(path):
            path = os.path.realpath(path)

        base_dir = self._base_dir or os.path.dirname(path)

        stack = Stack(is_nested=is_nested, name=name, assets=assets, origin_dir=os.path.dirname(path))

        template_dict = get_template_data(path)
        options = self._context.command_options_map
        resolved_stack = SamBaseProvider.get_resolved_template_dict(
            template_dict,
            SamLocalStackProvider.merge_parameter_overrides(
                options.get(PARAMETER_OVERRIDES), options.get(GLOBAL_PARAMETER_OVERRIDES)
            ),
            normalize_resource_metadata=False,
        )
        for key, value in resolved_stack.items():
            stack[key] = value

        resources_section = stack.get("Resources", DictSection())
        for resource in resources_section.section_items:
            resource_id = resource.item_id
            resource_type = resource.get("Type", None)
            properties = resource.get("Properties", {})
            package_type = properties.get("PackageType", ZIP)

            resource_assets: List[Asset] = []

            if resource_type in NESTED_STACKS_RESOURCES:
                nested_stack = self._extract_nested_stack(path, resource_id, properties, resource_type)
                resource.nested_stack = nested_stack

            if resource_type in RESOURCES_WITH_LOCAL_PATHS:
                for path_prop_name in RESOURCES_WITH_LOCAL_PATHS[resource_type]:
                    asset_path = jmespath.search(path_prop_name, properties)
                    if is_local_path(asset_path) and package_type == ZIP:
                        reference_path = base_dir if resource_type in BASE_DIR_RESOURCES else os.path.dirname(path)
                        asset_path = get_local_path(asset_path, reference_path)
                        asset = S3Asset(source_path=asset_path, source_property=path_prop_name)
                        resource_assets.append(asset)
                        stack.assets.append(asset)

            if resource_type in RESOURCES_WITH_IMAGE_COMPONENT:
                for path_prop_name in RESOURCES_WITH_IMAGE_COMPONENT[resource_type]:
                    asset_path = jmespath.search(path_prop_name, properties)
                    if asset_path and package_type == IMAGE:
                        asset = ImageAsset(source_local_image=asset_path, source_property=path_prop_name)
                        resource_assets.append(asset)
                        stack.assets.append(asset)

            resource.assets = resource_assets

        metadata_section = stack.get("Metadata", DictSection())
        for metadata in metadata_section.section_items:
            if not isinstance(metadata, DictSectionItem):
                continue
            metadata_type = metadata.item_id
            metadata_body = metadata.body
            metadata_assets: List[Asset] = []
            if metadata_type in METADATA_WITH_LOCAL_PATHS:
                for path_prop_name in METADATA_WITH_LOCAL_PATHS[metadata_type]:
                    asset_path = jmespath.search(path_prop_name, metadata_body)
                    asset = S3Asset(source_path=asset_path, source_property=path_prop_name)
                    metadata_assets.append(asset)
                    stack.assets.append(asset)

            metadata.assets = metadata_assets

        stack.extra_details[TEMPLATE_PATH_KEY] = path
        return stack

    def _extract_nested_stack(
        self, parent_stack_template_path, resource_id, properties, resource_type
    ) -> Optional[Stack]:
        if not properties:
            return None

        nested_stack_location_property_name = NESTED_STACKS_RESOURCES[resource_type]
        nested_stack_template_location = properties.get(nested_stack_location_property_name, None)

        if not is_local_path(nested_stack_template_location):
            return None

        if nested_stack_template_location.startswith("file://"):
            nested_stack_template_location = unquote(urlparse(nested_stack_template_location).path)
        else:
            nested_stack_template_location = SamLocalStackProvider.normalize_resource_path(
                os.path.dirname(parent_stack_template_path), nested_stack_template_location
            )
        nested_stack = self._build_stack(nested_stack_template_location, True, resource_id)
        nested_stack.extra_details[PARENT_STACK_TEMPLATE_PATH_KEY] = parent_stack_template_path
        return nested_stack

    @staticmethod
    def get_iac_file_patterns() -> List[str]:
        return ["template.yaml", "template.yml", "template.json"]



"""
The factory for returning the appropriate mapper and consumer
"""
from samcli.commands.list.json_consumer import StringConsumerJsonOutput
from samcli.commands.list.table_consumer import StringConsumerTableOutput
from samcli.lib.list.data_to_json_mapper import DataToJsonMapper
from samcli.lib.list.endpoints.endpoints_to_table_mapper import EndpointsToTableMapper
from samcli.lib.list.list_interfaces import Mapper, MapperConsumerFactoryInterface, ProducersEnum
from samcli.lib.list.mapper_consumer_container import MapperConsumerContainer
from samcli.lib.list.resources.resources_to_table_mapper import ResourcesToTableMapper
from samcli.lib.list.stack_outputs.stack_output_to_table_mapper import StackOutputToTableMapper


class MapperConsumerFactory(MapperConsumerFactoryInterface):
    """
    Factory class to create factory objects that map a given producer and output format to a mapper and a consumer
    """

    def create(self, producer: ProducersEnum, output: str) -> MapperConsumerContainer:
        """
        Creates a MapperConsumerContainer that contains the resulting mapper and consumer given
        the producer and output format

        Parameters
        ----------
        producer: ProducersEnum
            An enum representing the producers (stack-outputs, resources, or endpoints producer)
        output: str
            The output format, either json or table

        Returns
        -------
        container: MapperConsumerContainer
            A MapperConsumerContainer containing the resulting mapper and consumer to be used by the producer
        """
        if output == "json":
            data_to_json_mapper = DataToJsonMapper()
            json_consumer = StringConsumerJsonOutput()
            container = MapperConsumerContainer(data_to_json_mapper, json_consumer)
            return container
        table_mapper: Mapper
        table_consumer = StringConsumerTableOutput()
        if producer == ProducersEnum.STACK_OUTPUTS_PRODUCER:
            table_mapper = StackOutputToTableMapper()
        elif producer == ProducersEnum.RESOURCES_PRODUCER:
            table_mapper = ResourcesToTableMapper()
        elif producer == ProducersEnum.ENDPOINTS_PRODUCER:
            table_mapper = EndpointsToTableMapper()
        container = MapperConsumerContainer(table_mapper, table_consumer)
        return container



"""
Bootstrap's user's development environment by creating cloud resources required by SAM CLI
"""
import logging
from collections.abc import Collection
from typing import Dict, List, Optional, Union, cast

import boto3
import click
from botocore.config import Config
from botocore.exceptions import BotoCoreError, ClientError, NoCredentialsError, NoRegionError, ProfileNotFound

from samcli.commands.exceptions import CredentialsError, RegionError, UserException

LOG = logging.getLogger(__name__)


class ManagedStackError(UserException):
    def __init__(self, ex):
        self.ex = ex
        message_fmt = f"Failed to create managed resources: {ex}"
        super().__init__(message=message_fmt.format(ex=self.ex))


class StackOutput:
    def __init__(self, stack_output: List[Dict[str, str]]):
        self._stack_output: List[Dict[str, str]] = stack_output

    def get(self, key) -> Optional[str]:
        try:
            return next(o for o in self._stack_output if o.get("OutputKey") == key).get("OutputValue")
        except StopIteration:
            return None


def update_stack(
    region: Optional[str],
    stack_name: str,
    template_body: str,
    profile: Optional[str] = None,
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]] = None,
) -> StackOutput:
    """
    create or update a CloudFormation stack

    Parameters
    ----------
    region: str
        AWS region for the CloudFormation stack
    stack_name: str
        CloudFormation stack name
    template_body: str
        CloudFormation template's content
    profile: Optional[str]
        AWS named profile for the AWS account
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]]
        Values of template parameters, if any.

    Returns
    -------
    StackOutput:
        Stack output section(list of OutputKey, OutputValue pairs)
    """
    try:
        if profile:
            session = boto3.Session(profile_name=profile, region_name=region if region else None)
            cloudformation_client = session.client("cloudformation")
        else:
            cloudformation_client = boto3.client(
                "cloudformation", config=Config(region_name=region if region else None)
            )
    except ProfileNotFound as ex:
        raise CredentialsError(
            f"Error Setting Up Managed Stack Client: the provided AWS name profile '{profile}' is not found. "
            "please check the documentation for setting up a named profile: "
            "https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-profiles.html"
        ) from ex
    except NoCredentialsError as ex:
        raise CredentialsError(
            "Error Setting Up Managed Stack Client: Unable to resolve credentials for the AWS SDK for Python client. "
            "Please see their documentation for options to pass in credentials: "
            "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html"
        ) from ex
    except NoRegionError as ex:
        raise RegionError(
            "Error Setting Up Managed Stack Client: Unable to resolve a region. "
            "Please provide a region via the --region parameter or by the AWS_REGION environment variable."
        ) from ex
    return _create_or_update_stack(cloudformation_client, stack_name, template_body, parameter_overrides)


def manage_stack(
    region: Optional[str],
    stack_name: str,
    template_body: str,
    profile: Optional[str] = None,
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]] = None,
) -> StackOutput:
    """
    get or create a CloudFormation stack

    Parameters
    ----------
    region: str
        AWS region for the CloudFormation stack
    stack_name: str
        CloudFormation stack name
    template_body: str
        CloudFormation template's content
    profile: Optional[str]
        AWS named profile for the AWS account
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]]
        Values of template parameters, if any.

    Returns
    -------
    StackOutput:
        Stack output section(list of OutputKey, OutputValue pairs)
    """
    try:
        if profile:
            session = boto3.Session(profile_name=profile, region_name=region if region else None)
            cloudformation_client = session.client("cloudformation")
        else:
            cloudformation_client = boto3.client(
                "cloudformation", config=Config(region_name=region if region else None)
            )
    except ProfileNotFound as ex:
        raise CredentialsError(
            f"Error Setting Up Managed Stack Client: the provided AWS name profile '{profile}' is not found. "
            "please check the documentation for setting up a named profile: "
            "https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-profiles.html"
        ) from ex
    except NoCredentialsError as ex:
        raise CredentialsError(
            "Error Setting Up Managed Stack Client: Unable to resolve credentials for the AWS SDK for Python client. "
            "Please see their documentation for options to pass in credentials: "
            "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html"
        ) from ex
    except NoRegionError as ex:
        raise RegionError(
            "Error Setting Up Managed Stack Client: Unable to resolve a region. "
            "Please provide a region via the --region parameter or by the AWS_REGION environment variable."
        ) from ex
    return _create_or_get_stack(cloudformation_client, stack_name, template_body, parameter_overrides)


# Todo Add _update_stack to handle the case when the values of the stack parameter got changed
def _create_or_get_stack(
    cloudformation_client,
    stack_name: str,
    template_body: str,
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]] = None,
) -> StackOutput:
    try:
        ds_resp = cloudformation_client.describe_stacks(StackName=stack_name)
        stacks = ds_resp["Stacks"]
        stack = stacks[0]
        _check_sanity_of_stack(stack)
        stack_outputs = cast(List[Dict[str, str]], stack["Outputs"])
        return StackOutput(stack_outputs)
    except ClientError:
        LOG.debug("Managed S3 stack [%s] not found. Creating a new one.", stack_name)

    try:
        stack = _create_stack(
            cloudformation_client, stack_name, template_body, parameter_overrides
        )  # exceptions are not captured from subcommands
        _check_sanity_of_stack(stack)
        stack_outputs = cast(List[Dict[str, str]], stack["Outputs"])
        return StackOutput(stack_outputs)
    except (ClientError, BotoCoreError) as ex:
        LOG.debug("Failed to create managed resources", exc_info=ex)
        raise ManagedStackError(str(ex)) from ex


def _create_or_update_stack(
    cloudformation_client,
    stack_name: str,
    template_body: str,
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]] = None,
) -> StackOutput:
    try:
        cloudformation_client.describe_stacks(StackName=stack_name)
        stack = _update_stack(cloudformation_client, stack_name, template_body, parameter_overrides)
        _check_sanity_of_stack(stack)
        stack_outputs = cast(List[Dict[str, str]], stack["Outputs"])
        return StackOutput(stack_outputs)
    except ClientError:
        LOG.debug("Managed S3 stack [%s] not found. Creating a new one.", stack_name)

    try:
        stack = _create_stack(
            cloudformation_client, stack_name, template_body, parameter_overrides
        )  # exceptions are not captured from subcommands
        _check_sanity_of_stack(stack)
        stack_outputs = cast(List[Dict[str, str]], stack["Outputs"])
        return StackOutput(stack_outputs)
    except (ClientError, BotoCoreError) as ex:
        LOG.debug("Failed to create managed resources", exc_info=ex)
        raise ManagedStackError(str(ex)) from ex


def _check_sanity_of_stack(stack):
    stack_name = stack.get("StackName")
    tags = stack.get("Tags", None)
    outputs = stack.get("Outputs", None)

    # For some edge cases, stack could be in invalid state
    # Check if stack information contains the Tags and Outputs as we expected
    if tags is None or outputs is None:
        stack_state = stack.get("StackStatus", None)
        msg = (
            f"Stack {stack_name} is missing Tags and/or Outputs information and therefore not in a "
            f"healthy state (Current state:{stack_state}). Failing as the stack was likely not created "
            f"by the AWS SAM CLI"
        )
        raise UserException(msg)

    # Sanity check for non-none stack? Sanity check for tag?
    try:
        sam_cli_tag = next(t for t in tags if t["Key"] == "ManagedStackSource")
        if not sam_cli_tag["Value"] == "AwsSamCli":
            msg = (
                "Stack "
                + stack_name
                + " ManagedStackSource tag shows "
                + sam_cli_tag["Value"]
                + " which does not match the AWS SAM CLI generated tag value of AwsSamCli. "
                "Failing as the stack was likely not created by the AWS SAM CLI."
            )
            raise UserException(msg)
    except StopIteration as ex:
        msg = (
            "Stack  " + stack_name + " exists, but the ManagedStackSource tag is missing. "
            "Failing as the stack was likely not created by the AWS SAM CLI."
        )
        raise UserException(msg) from ex


def _create_stack(
    cloudformation_client,
    stack_name: str,
    template_body: str,
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]] = None,
):
    click.echo("\tCreating the required resources...")
    change_set_name = "InitialCreation"
    parameters = _generate_stack_parameters(parameter_overrides)
    change_set_resp = cloudformation_client.create_change_set(
        StackName=stack_name,
        TemplateBody=template_body,
        Tags=[{"Key": "ManagedStackSource", "Value": "AwsSamCli"}],
        ChangeSetType="CREATE",
        ChangeSetName=change_set_name,  # this must be unique for the stack, but we only create so that's fine
        Capabilities=["CAPABILITY_IAM"],
        Parameters=parameters,
    )
    stack_id = change_set_resp["StackId"]
    change_waiter = cloudformation_client.get_waiter("change_set_create_complete")
    change_waiter.wait(
        ChangeSetName=change_set_name, StackName=stack_name, WaiterConfig={"Delay": 15, "MaxAttempts": 60}
    )
    cloudformation_client.execute_change_set(ChangeSetName=change_set_name, StackName=stack_name)
    stack_waiter = cloudformation_client.get_waiter("stack_create_complete")
    stack_waiter.wait(StackName=stack_id, WaiterConfig={"Delay": 15, "MaxAttempts": 60})
    ds_resp = cloudformation_client.describe_stacks(StackName=stack_name)
    stacks = ds_resp["Stacks"]
    click.echo("\tSuccessfully created!")
    return stacks[0]


def _update_stack(
    cloudformation_client,
    stack_name: str,
    template_body: str,
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]] = None,
):
    click.echo("\tUpdating the required resources...")
    parameters = _generate_stack_parameters(parameter_overrides)
    us_resp = cloudformation_client.update_stack(
        StackName=stack_name,
        TemplateBody=template_body,
        Tags=[{"Key": "ManagedStackSource", "Value": "AwsSamCli"}],
        Capabilities=["CAPABILITY_IAM", "CAPABILITY_AUTO_EXPAND"],
        Parameters=parameters,
    )
    stack_id = us_resp["StackId"]
    stack_waiter = cloudformation_client.get_waiter("stack_update_complete")
    stack_waiter.wait(StackName=stack_id, WaiterConfig={"Delay": 15, "MaxAttempts": 60})
    ds_resp = cloudformation_client.describe_stacks(StackName=stack_name)
    stacks = ds_resp["Stacks"]
    click.echo("\tSuccessfully updated!")
    return stacks[0]


def _generate_stack_parameters(
    parameter_overrides: Optional[Dict[str, Union[str, List[str]]]] = None
) -> List[Dict[str, str]]:
    parameters = []
    if parameter_overrides:
        for key, value in parameter_overrides.items():
            if isinstance(value, Collection) and not isinstance(value, str):
                # Assumption: values don't include commas or spaces. Need to refactor to handle such a case if needed.
                value = ",".join(value)
            parameters.append({"ParameterKey": key, "ParameterValue": value})
    return parameters



"""
Image artifacts based utilities
"""
import docker
from docker.errors import APIError, NullResource

from samcli.commands.package.exceptions import DockerGetLocalImageFailedError
from samcli.lib.package.utils import is_ecr_url

SHA_CHECKSUM_TRUNCATION_LENGTH = 12


class NonLocalImageException(Exception):
    pass


class NoImageFoundException(Exception):
    pass


def tag_translation(image, docker_image_id=None, gen_tag="latest"):
    """
    Translates a given local image structure such as `helloworld:v1` into
    just a tag structure such as `helloworld-v1` , this tag will then be applied
    to an image that is to be uploaded to a remote registry.

    :param docker_image_id: an Id associated with a docker image.
    :param image: an image referenceable by docker locally.
    :param gen_tag: tag to be generated if the image did not have a tag assigned to it.
    :return: fully qualified tag.
    """
    # NOTE(sriram-mv): assumption on tag structure, needs vetting.
    if is_ecr_url(image):
        raise NonLocalImageException(f"{image} is a non-local image")

    if not docker_image_id:
        try:
            docker_client = docker.from_env()
            docker_image_id = docker_client.images.get(image).id
        except APIError as ex:
            raise DockerGetLocalImageFailedError(str(ex)) from ex
        except NullResource as ex:
            raise NoImageFoundException(str(ex)) from ex

    # NOTE(sriram-mv): Checksum truncation Length is set to 12
    _id = docker_image_id.split(":")[1][:SHA_CHECKSUM_TRUNCATION_LENGTH]
    if ":" in image:
        name, tag = image.split(":")
    else:
        name = image
        tag = None
    _tag = tag if tag else gen_tag
    return f"{name}-{_id}-{_tag}"



"""
Utilities for table pretty printing using click
"""
import shutil
import textwrap
from functools import wraps
from itertools import count, zip_longest
from typing import Sized

import click

MIN_OFFSET = 20


def pprint_column_names(
    format_string, format_kwargs, margin=None, table_header=None, color="yellow", display_sleep=False
):
    """

    :param format_string: format string to be used that has the strings, minimum width to be replaced
    :param format_kwargs: dictionary that is supplied to the format_string to format the string
    :param margin: margin that is to be reduced from column width for columnar text.
    :param table_header: Supplied table header
    :param color: color supplied for table headers and column names.
    :param display_sleep: flag to format table_header to include deployer's client_sleep
    :return: boilerplate table string
    """

    min_width = 100
    min_margin = 2

    def pprint_wrap(func):
        # Calculate terminal width, number of columns in the table
        width, _ = shutil.get_terminal_size()
        # For UX purposes, set a minimum width for the table to be usable
        # and usable_width keeps margins in mind.
        width = max(width, min_width)

        total_args = len(format_kwargs)
        if not total_args:
            raise ValueError("Number of arguments supplied should be > 0 , format_kwargs: {}".format(format_kwargs))

        # Get width to be a usable number so that we can equally divide the space for all the columns.
        # Can be refactored, to allow for modularity in the shaping of the columns.
        width = width - (width % total_args)
        usable_width_no_margin = int(width) - 1
        usable_width = int((usable_width_no_margin - (margin if margin else min_margin)))
        if total_args > int(usable_width / 2):
            raise ValueError("Total number of columns exceed available width")
        width_per_column = int(usable_width / total_args)

        # The final column should not roll over into the next line
        final_arg_width = width_per_column - 1

        # the format string contains minimumwidth that need to be set.
        # eg: "{a:{0}}} {b:<{1}}} {c:{2}}}"
        format_args = [width_per_column for _ in range(total_args - 1)]
        format_args.extend([final_arg_width])

        # format arguments are now ready for setting minimumwidth

        @wraps(func)
        def wrap(*args, **kwargs):
            # The table is setup with the column names, format_string contains the column names.
            if table_header:
                click.secho(
                    "\n" + table_header.format(args[0].client_sleep) if display_sleep else table_header, bold=True
                )
            click.secho("-" * usable_width, fg=color)
            click.secho(format_string.format(*format_args, **format_kwargs), fg=color)
            click.secho("-" * usable_width, fg=color)
            # format_args which have the minimumwidth set per {} in the format_string is passed to the function
            # which this decorator wraps, so that the function has access to the correct format_args
            kwargs["format_args"] = format_args
            kwargs["width"] = width_per_column
            kwargs["margin"] = margin if margin else min_margin
            result = func(*args, **kwargs)
            # Complete the table
            click.secho("-" * usable_width + "\n", fg=color)
            return result

        return wrap

    return pprint_wrap


def wrapped_text_generator(texts, width, margin, **textwrap_kwargs):
    """

    Return a generator where the contents are wrapped text to a specified width.

    :param texts: list of text that needs to be wrapped at specified width
    :param width: width of the text to be wrapped
    :param margin: margin to be reduced from width for cleaner UX
    :param textwrap_kwargs: keyword arguments that are passed to textwrap.wrap
    :return: generator of wrapped text
    :rtype: Iterator[str]
    """
    for text in texts:
        yield textwrap.wrap(text, width=width - margin, **textwrap_kwargs)


def pprint_columns(columns, width, margin, format_string, format_args, columns_dict, color="yellow", **textwrap_kwargs):
    """

    Print columns based on list of columnar text, associated formatting string and associated format arguments.

    :param columns: list of columnnar text that go into columns as specified by the format_string
    :param width: width of the text to be wrapped
    :param margin: margin to be reduced from width for cleaner UX
    :param format_string: A format string that has both width and text specifiers set.
    :param format_args: list of offset specifiers
    :param columns_dict: arguments dictionary that have dummy values per column
    :param color: color supplied for rows within the table.
    :param textwrap_kwargs: keyword arguments that are passed to textwrap.wrap
    """
    for columns_text in zip_longest(*wrapped_text_generator(columns, width, margin, **textwrap_kwargs), fillvalue=""):
        counter = count()
        # Generate columnar data that correspond to the column names and update them.
        for k, _ in columns_dict.items():
            columns_dict[k] = columns_text[next(counter)]

        click.secho(format_string.format(*format_args, **columns_dict), fg=color)


def newline_per_item(iterable: Sized, counter: int) -> None:
    """
    Adds a new line based on the index of a given iterable
    Parameters
    ----------
    iterable: Any iterable that implements __len__
    counter: Current index within the iterable

    Returns
    -------

    """
    if counter < len(iterable) - 1:
        click.echo(message="", nl=True)



"""
Utilities for checking authorization of certain resource types
"""
import logging
from typing import List, Tuple

from samcli.commands.local.lib.swagger.reader import SwaggerReader
from samcli.lib.providers.provider import Stack
from samcli.lib.providers.sam_function_provider import SamFunctionProvider

LOG = logging.getLogger(__name__)


def auth_per_resource(stacks: List[Stack]):
    """
    Check if authentication has been set for the function resources defined in the template that have `Api` Event type
    or the function property FunctionUrlConfig.

    Parameters
    ----------
    stacks: List[Stack]
        The list of stacks where resources are looked for

    Returns
    -------

    List of tuples per function resource that have the `Api` or `HttpApi` event types, that describes the resource
    (function logical_id - event type or function resource logical_id and description - FURL
    and if authorization is required per resource.

    """
    _auth_per_resource: List[Tuple[str, bool]] = []

    sam_function_provider = SamFunctionProvider(stacks, ignore_code_extraction_warnings=True)

    for sam_function in sam_function_provider.get_all():
        # Only check for auth if there are function events defined or function url Configuration is defined.
        if sam_function.events:
            _auth_resource_event(sam_function_provider, sam_function, _auth_per_resource)

        if sam_function.function_url_config:
            authorization_type = sam_function.function_url_config.get("AuthType")
            function_resource_name = f"{sam_function.name} Function Url"
            _auth_per_resource.append((function_resource_name, bool(authorization_type != "NONE")))

    return _auth_per_resource


def _auth_resource_event(sam_function_provider: SamFunctionProvider, sam_function, auth_resource_list):
    """

    Parameters
    ----------
    sam_function_provider: SamFunctionProvider
    sam_function: Current function which has all intrinsics resolved.
    auth_resource_list: List of tuples with function name and auth. eg: [("Name", True)]

    Returns
    -------

    """
    for event in sam_function.events.values():
        for event_type, identifier in [("Api", "RestApiId"), ("HttpApi", "ApiId")]:
            if event.get("Type") == event_type:
                # Is there any auth defined directly on the function resource?
                # NOTE(sriram-mv): How do we check this in more detail?
                # Currently just checking for presence of Auth, not the details.
                # https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/sam-property-function-apifunctionauth.html
                # https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/sam-property-function-httpapifunctionauth.html
                if event.get("Properties", {}).get("Auth", False):
                    auth_resource_list.append((sam_function.name, True))
                # Is there any auth defined on the referred http api or serverless api through the `id` construct?
                elif _auth_id(
                    # use the resources containing the sam_function
                    sam_function_provider.get_resources_by_stack_path(sam_function.stack_path),
                    event.get("Properties", {}),
                    identifier,
                ):
                    auth_resource_list.append((sam_function.name, True))
                else:
                    auth_resource_list.append((sam_function.name, False))


def _auth_id(resources_dict, event_properties, identifier):
    """

    Parameters
    ----------
    resources_dict: dict
        Resolved resources defined in the SAM Template
    event_properties: dict
        Properties of given event supplied to a function resource
    identifier: str
        Id: `ApiId` or `RestApiId`

    Returns
    -------
    bool
        Returns if the given identifier under the event properties maps to a resource and has authorization enabled.

    """
    resource_name = event_properties.get(identifier, "")
    api_resource = resources_dict.get(resource_name, {})
    return any(
        [
            api_resource.get("Properties", {}).get("Auth", False),
            _auth_definition_body_and_uri(
                definition_body=api_resource.get("Properties", {}).get("DefinitionBody", {}),
                definition_uri=api_resource.get("Properties", {}).get("DefinitionUri", None),
            ),
        ]
    )


def _auth_definition_body_and_uri(definition_body, definition_uri):
    """

    Parameters
    ----------
    definition_body: dict
        inline definition body defined in the template
    definition_uri: string
        Either an s3 url or a local path to a definition uri

    Returns
    -------
    bool
        Is security defined on the swagger or not?


    """

    reader = SwaggerReader(definition_body=definition_body, definition_uri=definition_uri)
    swagger = reader.read()
    _auths = []
    if not swagger:
        swagger = {}
    # NOTE(sriram-mv): Authorization and Authentication is indicated by the `security` scheme.
    # https://swagger.io/docs/specification/authentication/
    for _, verb in swagger.get("paths", {}).items():
        for _property in verb.values():
            # If there are instrinsics in play, they may not be resolved yet.
            if isinstance(_property, dict):
                _auths.append(bool(_property.get("security", False)))

    _auths.append(bool(swagger.get("security", False)))

    if swagger:
        LOG.debug("Auth checks done on swagger are not exhaustive!")

    # This is not an exhaustive check, but to check if there is some form of security setup.
    return any(_auths)



"""
Common context class to inherit from for sam list sub-commands
"""
from samcli.lib.utils.boto_utils import get_boto_client_provider_with_config


class ListContext:
    def __init__(self):
        self.cloudformation_client = None
        self.client_provider = None
        self.region = None
        self.profile = None

    def init_clients(self) -> None:
        """
        Initialize the clients being used by sam list.
        """
        from boto3 import Session

        if not self.region:
            session = Session()
            self.region = session.region_name

        client_provider = get_boto_client_provider_with_config(region=self.region, profile=self.profile)
        self.client_provider = client_provider
        self.cloudformation_client = client_provider("cloudformation")



"""
Common CLI options for invoke command
"""
from pathlib import Path

import click

from samcli.commands._utils.options import docker_click_options, parameter_override_click_option, template_click_option
from samcli.commands.local.cli_common.invoke_context import ContainersInitializationMode


def get_application_dir():
    """

    Returns
    -------
    Path
        Path representing the application config directory
    """
    # TODO: Get the config directory directly from `GlobalConfig`
    return Path(click.get_app_dir("AWS SAM", force_posix=True))


def get_default_layer_cache_dir():
    """
    Default the layer cache directory

    Returns
    -------
    str
        String representing the layer cache directory
    """
    layer_cache_dir = get_application_dir().joinpath("layers-pkg")

    return str(layer_cache_dir)


def local_common_options(f):
    """
    Common CLI options shared by "local invoke", "local start-api", and "local start-lambda" commands

    :param f: Callback passed by Click
    """
    local_options = [
        click.option(
            "--shutdown",
            is_flag=True,
            default=False,
            help="If set, will emulate a shutdown event after the invoke completes, "
            "in order to test extension handling of shutdown behavior.",
        ),
        click.option(
            "--container-host",
            default="localhost",
            show_default=True,
            help="Host of locally emulated Lambda container. "
            "This option is useful when the container runs on a different host than SAM CLI. "
            "For example, if you want to run SAM CLI in a Docker container on macOS, "
            "use this option with host.docker.internal",
        ),
        click.option(
            "--container-host-interface",
            default="127.0.0.1",
            show_default=True,
            help="IP address of the host network interface that container ports should bind to. "
            "Use 0.0.0.0 to bind to all interfaces.",
        ),
        click.option(
            "--invoke-image",
            "-ii",
            default=None,
            required=False,
            multiple=True,
            help="Container image URIs for invoking functions or starting api and function. "
            "You can specify the image URI used for the local function invocation "
            "(--invoke-image public.ecr.aws/sam/build-nodejs14.x:latest). "
            "You can specify for each individual function with "
            "(--invoke-image Function1=public.ecr.aws/sam/build-nodejs14.x:latest). "
            "If a function does not have invoke image specified, the default SAM CLI "
            "emulation image will be used.",
        ),
    ]

    # Reverse the list to maintain ordering of options in help text printed with --help
    for option in reversed(local_options):
        option(f)

    return f


def service_common_options(port):
    """
    Construct common CLI Options that are shared for service related commands ('start-api' and 'start_lambda')

    Parameters
    ----------
    port
        The port number to listen to
    """

    def construct_options(f):
        """
        Common CLI Options that are shared for service related commands ('start-api' and 'start_lambda')

        Parameters
        ----------
        f function
            Callback passed by Click
        port int
            port number to use

        Returns
        -------
        function
            The callback function
        """
        service_options = [
            click.option(
                "--host", default="127.0.0.1", help="Local hostname or IP address to bind to (default: '127.0.0.1')"
            ),
            click.option(
                "--port", "-p", default=port, help="Local port number to listen on (default: '{}')".format(str(port))
            ),
        ]

        # Reverse the list to maintain ordering of options in help text printed with --help
        for option in reversed(service_options):
            option(f)

        return f

    return construct_options


def invoke_common_options(f):
    """
    Common CLI options shared by "local invoke" and "local start-api" commands

    :param f: Callback passed by Click
    """

    invoke_options = (
        [
            template_click_option(),
            click.option(
                "--env-vars",
                "-n",
                type=click.Path(exists=True),
                help="JSON file containing values for Lambda function's environment variables.",
            ),
            parameter_override_click_option(),
            click.option(
                "--debug-port",
                "-d",
                help="When specified, Lambda function container will start in debug mode and will expose this "
                "port on localhost.",
                envvar="SAM_DEBUG_PORT",
                type=click.INT,
                multiple=True,
            ),
            click.option(
                "--debugger-path", help="Host path to a debugger that will be mounted into the Lambda container."
            ),
            click.option(
                "--debug-args", help="Additional arguments to be passed to the debugger.", envvar="DEBUGGER_ARGS"
            ),
            click.option(
                "--container-env-vars",
                type=click.Path(exists=True),
                help="JSON file containing environment variables to be set within the container environment",
            ),
            click.option(
                "--docker-volume-basedir",
                "-v",
                envvar="SAM_DOCKER_VOLUME_BASEDIR",
                help="Specifies the location basedir where the SAM file exists. If the Docker is running on "
                "a remote machine, you must mount the path where the SAM file exists on the docker machine "
                "and modify this value to match the remote machine.",
            ),
            click.option("--log-file", "-l", help="logfile to send runtime logs to."),
            click.option(
                "--layer-cache-basedir",
                type=click.Path(exists=False, file_okay=False),
                envvar="SAM_LAYER_CACHE_BASEDIR",
                help="Specifies the location basedir where the Layers your template uses will be downloaded to.",
                default=get_default_layer_cache_dir(),
            ),
        ]
        + docker_click_options()
        + [
            click.option(
                "--force-image-build",
                is_flag=True,
                help="Specify whether CLI should rebuild the image used for invoking functions with layers.",
                envvar="SAM_FORCE_IMAGE_BUILD",
                default=False,
            )
        ]
    )

    # Reverse the list to maintain ordering of options in help text printed with --help
    for option in reversed(invoke_options):
        option(f)

    return f


def warm_containers_common_options(f):
    """
    Warm containers related CLI options shared by "local start-api" and "local start_lambda" commands

    :param f: Callback passed by Click
    """

    warm_containers_options = [
        click.option(
            "--warm-containers",
            help="""
            \b
            Optional. Specifies how AWS SAM CLI manages 
            containers for each function.
            Two modes are available:
            EAGER: Containers for all functions are 
            loaded at startup and persist between 
            invocations.
            LAZY:  Containers are only loaded when each 
            function is first invoked. Those containers 
            persist for additional invocations.
            """,
            type=click.Choice(ContainersInitializationMode.__members__, case_sensitive=False),
        ),
        click.option(
            "--debug-function",
            help="Optional. Specifies the Lambda Function logicalId to apply debug options to when"
            " --warm-containers is specified.  This parameter applies to --debug-port, --debugger-path,"
            " and --debug-args.",
            type=click.STRING,
            multiple=False,
        ),
    ]

    # Reverse the list to maintain ordering of options in help text printed with --help
    for option in reversed(warm_containers_options):
        option(f)

    return f



"""
Classes which will change permissions on a ZipInfo object
"""
import platform
import zipfile


class PermissionMapper:
    """
    Base Mapper with an empty apply method.
    """

    def apply(self, zip_info: zipfile.ZipInfo):
        pass


class WindowsFilePermissionPermissionMapper(PermissionMapper):
    """
    Windows specific Permission Mapper onto a zipfile.ZipInfo object to
    set explicit permissions onto `external_attr` attribute if the given object
    is of type `file`
    """

    def __init__(self, permissions: int):
        """

        Parameters
        ----------
        permissions: int
            Base permissions.
        """
        self.permissions = permissions

    def apply(self, zip_info: zipfile.ZipInfo):
        """

        Parameters
        ----------
        zip_info: zipfile.ZipInfo
            object of zipfile.ZipInfo

        Returns
        -------
        zipfile.ZipInfo
            modified object with `external_attr` set to member permissions.

        """
        if platform.system().lower() == "windows" and not zip_info.is_dir():
            zip_info.external_attr = self.permissions << 16
        return zip_info


class WindowsDirPermissionPermissionMapper(PermissionMapper):
    """
    Windows specific Permission Mapper onto a zipfile.ZipInfo object to
    set explicit permissions onto `external_attr` attribute if the given object
    is of type `directory`
    """

    def __init__(self, permissions: int):
        """

        Parameters
        ----------
        permissions: int
            Base permissions.
        """
        self.permissions = permissions

    def apply(self, zip_info: zipfile.ZipInfo):
        """

        Parameters
        ----------
        zip_info: zipfile.ZipInfo
            object of zipfile.ZipInfo

        Returns
        -------
        zipfile.ZipInfo
            modified object with `external_attr` set to member permissions.

        """
        if platform.system().lower() == "windows" and zip_info.is_dir():
            zip_info.external_attr = self.permissions << 16
        return zip_info


class AdditiveDirPermissionPermissionMapper(PermissionMapper):
    """
    Additive Permission Mapper onto a zipfile.ZipInfo object to
    set additive permissions onto `external_attr` attribute if the given object
    is of type `directory`
    """

    def __init__(self, permissions: int):
        """

        Parameters
        ----------
        permissions: int
            Base permissions.
        """
        self.permissions = permissions

    def apply(self, zip_info: zipfile.ZipInfo):
        """

        Parameters
        ----------
        zip_info: zipfile.ZipInfo
            object of zipfile.ZipInfo

        Returns
        -------
        zipfile.ZipInfo
            modified object with `external_attr` added with member permissions
            if conditions are satisfied.

        """
        if zip_info.is_dir():
            zip_info.external_attr |= self.permissions << 16
        return zip_info


class AdditiveFilePermissionPermissionMapper(PermissionMapper):
    """
    Additive Permission Mapper onto a zipfile.ZipInfo object to
    set additive permissions onto `external_attr` attribute if the given object
    is of type `file`
    """

    def __init__(self, permissions: int):
        """

        Parameters
        ----------
        permissions: int
            Base permissions.
        """
        self.permissions = permissions

    def apply(self, zip_info: zipfile.ZipInfo):
        """

        Parameters
        ----------
        zip_info: zipfile.ZipInfo
            object of zipfile.ZipInfo

        Returns
        -------
        zipfile.ZipInfo
            modified object with `external_attr` added with member permissions
            if conditions are satisfied.

        """
        if not zip_info.is_dir():
            zip_info.external_attr |= self.permissions << 16
        return zip_info



"""
Implementation of the resources to table mapper
"""
from collections import OrderedDict
from typing import Any, Dict

from samcli.lib.list.list_interfaces import Mapper


class ResourcesToTableMapper(Mapper):
    """
    Mapper class for mapping resources data for table output
    """

    def map(self, data: list) -> Dict[Any, Any]:
        """
        Maps data to the format needed for consumption by the table consumer

        Parameters
        ----------
        data: list
            List of dictionaries containing the entries of the resources data

        Returns
        -------
        table_data: Dict[Any, Any]
            Dictionary containing the information and data needed for the table
            consumer to output the data in table format
        """
        entry_list = []
        for resource in data:
            entry_list.append(
                [
                    resource.get("LogicalResourceId", "-"),
                    resource.get("PhysicalResourceId", "-"),
                ]
            )
        table_data = {
            "format_string": "{Logical ID:<{0}} {Physical ID:<{1}}",
            "format_args": OrderedDict(
                {
                    "Logical ID": "Logical ID",
                    "Physical ID": "Physical ID",
                }
            ),
            "table_name": "Resources",
            "data": entry_list,
        }
        return table_data



"""
Context object used by sync command
"""
import logging
import threading
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional, cast

import tomlkit
from tomlkit.items import Item
from tomlkit.toml_document import TOMLDocument

from samcli.lib.build.build_graph import DEFAULT_DEPENDENCIES_DIR
from samcli.lib.utils.osutils import rmtree_if_exists

LOG = logging.getLogger(__name__)


DEFAULT_SYNC_STATE_FILE_NAME = "sync.toml"

SYNC_STATE = "sync_state"
RESOURCE_SYNC_STATES = "resource_sync_states"
HASH = "hash"
SYNC_TIME = "sync_time"
DEPENDENCY_LAYER = "dependency_layer"

# global lock for writing to file
_lock = threading.Lock()


@dataclass
class ResourceSyncState:
    hash_value: str
    sync_time: datetime


@dataclass
class SyncState:
    dependency_layer: bool
    resource_sync_states: Dict[str, ResourceSyncState]

    def update_resource_sync_state(self, resource_id: str, hash_value: str):
        self.resource_sync_states[resource_id] = ResourceSyncState(hash_value, datetime.utcnow())


def _sync_state_to_toml_document(sync_state: SyncState) -> TOMLDocument:
    """
    Writes the sync state information to the TOML file.

    Parameters
    -------
    sync_state: SyncState
        The SyncState to cache the information in the TOML file

    Returns
    -------
    TOMLDocument
        Object which will be dumped to the TOML file
    """
    sync_state_toml_table = tomlkit.table()
    sync_state_toml_table[DEPENDENCY_LAYER] = sync_state.dependency_layer

    resource_sync_states_toml_table = tomlkit.table()
    for resource_id in sync_state.resource_sync_states:
        resource_sync_state = sync_state.resource_sync_states[resource_id]

        resource_sync_state_toml_table = tomlkit.table()

        resource_sync_state_toml_table[HASH] = resource_sync_state.hash_value
        resource_sync_state_toml_table[SYNC_TIME] = resource_sync_state.sync_time.isoformat()

        # For Nested stack resources, replace "/" with "-"
        resource_id_toml = resource_id.replace("/", "-")
        resource_sync_states_toml_table[resource_id_toml] = resource_sync_state_toml_table

    toml_document = tomlkit.document()
    toml_document.add((tomlkit.comment("This file is auto generated by SAM CLI sync command")))
    toml_document.add(SYNC_STATE, cast(Item, sync_state_toml_table))
    toml_document.add(RESOURCE_SYNC_STATES, cast(Item, resource_sync_states_toml_table))

    return toml_document


def _toml_document_to_sync_state(toml_document: Dict) -> Optional[SyncState]:
    """
    Reads the cached information from the provided toml_document.

    Parameters
    -------
    toml_document: SyncState
        The toml document to read the information from

    """
    if not toml_document:
        return None

    sync_state_toml_table = toml_document.get(SYNC_STATE)
    resource_sync_states_toml_table = toml_document.get(RESOURCE_SYNC_STATES, {})

    # If no info in toml file
    if not (sync_state_toml_table or resource_sync_states_toml_table):
        return None

    resource_sync_states = dict()
    if resource_sync_states_toml_table:
        for resource_id in resource_sync_states_toml_table:

            resource_sync_state_toml_table = resource_sync_states_toml_table.get(resource_id)
            resource_sync_state = ResourceSyncState(
                resource_sync_state_toml_table.get(HASH),
                datetime.fromisoformat(resource_sync_state_toml_table.get(SYNC_TIME)),
            )

            # For Nested stack resources, replace "-" with "/"
            resource_sync_state_resource_id = resource_id.replace("-", "/")
            resource_sync_states[resource_sync_state_resource_id] = resource_sync_state

    dependency_layer = False
    if sync_state_toml_table:
        dependency_layer = sync_state_toml_table.get(DEPENDENCY_LAYER)
    sync_state = SyncState(dependency_layer, resource_sync_states)

    return sync_state


class SyncContext:

    _current_state: SyncState
    _previous_state: Optional[SyncState]
    _build_dir: Path
    _cache_dir: Path
    _file_path: Path

    def __init__(self, dependency_layer: bool, build_dir: str, cache_dir: str):
        self._current_state = SyncState(dependency_layer, dict())
        self._previous_state = None
        self._build_dir = Path(build_dir)
        self._cache_dir = Path(cache_dir)
        self._file_path = Path(build_dir).parent.joinpath(DEFAULT_SYNC_STATE_FILE_NAME)

    def __enter__(self) -> "SyncContext":
        with _lock:
            self._read()
        LOG.debug(
            "Entering sync context, previous state: %s, current state: %s", self._previous_state, self._current_state
        )

        # if adl parameter is changed between sam sync runs, cleanup build, cache and dependencies folders
        if self._previous_state and self._previous_state.dependency_layer != self._current_state.dependency_layer:
            self._cleanup_build_folders()

        return self

    def __exit__(self, *args) -> None:
        with _lock:
            self._write()

    def update_resource_sync_state(self, resource_id: str, hash_value: str) -> None:
        """
        Updates the sync_state information for the provided resource_id
        to be stored in the TOML file.

        Parameters
        -------
        resource_id: str
            The resource identifier of the resource
        hash_value: str
            The logical ID identifier of the resource
        """
        with _lock:
            LOG.debug("Updating resource_sync_state for resource %s with hash %s", resource_id, hash_value)
            self._current_state.update_resource_sync_state(resource_id, hash_value)
            self._write()

    def get_resource_latest_sync_hash(self, resource_id: str) -> Optional[str]:
        """
        Returns the latest hash from resource_sync_state if this information was
        cached for the provided resource_id.

        Parameters
        -------
        resource_id: str
            The resource identifier of the resource

        Returns
        -------
        Optional[str]
            The hash of the resource stored in resource_sync_state if it exists
        """
        with _lock:
            resource_sync_state = self._current_state.resource_sync_states.get(resource_id)
            if not resource_sync_state:
                LOG.debug("No latest hash found for resource %s", resource_id)
                return None
            LOG.debug(
                "Latest resource_sync_state hash %s found for resource %s", resource_id, resource_sync_state.hash_value
            )
            return resource_sync_state.hash_value

    def _write(self) -> None:
        with open(self._file_path, "w+") as file:
            file.write(tomlkit.dumps(_sync_state_to_toml_document(self._current_state)))

    def _read(self) -> None:
        try:
            with open(self._file_path) as file:
                toml_document = cast(Dict, tomlkit.loads(file.read()))
            self._previous_state = _toml_document_to_sync_state(toml_document)
            if self._previous_state:
                self._current_state.resource_sync_states = self._previous_state.resource_sync_states
        except OSError:
            LOG.debug("Missing previous sync state, will create a new file at the end of this execution")

    def _cleanup_build_folders(self) -> None:
        """
        Cleans up build, cache and dependencies folders for clean start of the next session
        """
        LOG.debug("Cleaning up build directory %s", self._build_dir)
        rmtree_if_exists(self._build_dir)

        LOG.debug("Cleaning up cache directory %s", self._cache_dir)
        rmtree_if_exists(self._cache_dir)

        dependencies_dir = Path(DEFAULT_DEPENDENCIES_DIR)
        LOG.debug("Cleaning up dependencies directory: %s", dependencies_dir)
        rmtree_if_exists(dependencies_dir)



"""
Implementation of custom click parameter types
"""

import json
import re
from json import JSONDecodeError

import click

from samcli.lib.package.ecr_utils import is_ecr_url

PARAM_AND_METADATA_KEY_REGEX = """([A-Za-z0-9\\"\']+)"""


def _generate_match_regex(match_pattern, delim):
    """
    Creates a regex string based on a match pattern (also a regex) that is to be
    run on a string (which may contain escaped quotes) that is separated by delimiters.

    Parameters
    ----------
    match_pattern: (str) regex pattern to match
    delim: (str) delimiter that is respected when identifying matching groups with generated regex.

    Returns
    -------
    str: regex expression

    """

    # Non capturing groups reduces duplicates in groups, but does not reduce matches.
    return (
        f"""(\\"(?:\\\\{match_pattern}|[^\\"\\\\]+)*\\"|"""
        + f"""\'(?:\\\\{match_pattern}|[^\'\\\\]+)*\'|"""
        + f"""(?:\\\\{match_pattern}|[^{delim}\\"\\\\]+)+)"""
    )


def _unquote_wrapped_quotes(value):
    r"""
    Removes wrapping single or double quotes and unescapes '\ ', '\"' and '\''.

    Parameters
    ----------
    value : str
        Input to unquote

    Returns
    -------
    Unquoted string
    """

    if value and ((value[0] == value[-1] == '"') or (value[0] == value[-1] == "'")):
        value = value[1:-1]

    return value.replace("\\ ", " ").replace('\\"', '"').replace("\\'", "'")


class CfnParameterOverridesType(click.ParamType):
    """
    Custom Click options type to accept values for CloudFormation template parameters. You can pass values for
    parameters as "ParameterKey=KeyPairName,ParameterValue=MyKey ParameterKey=InstanceType,ParameterValue=t1.micro"
    """

    __EXAMPLE_1 = "ParameterKey=KeyPairName,ParameterValue=MyKey ParameterKey=InstanceType,ParameterValue=t1.micro"
    __EXAMPLE_2 = "KeyPairName=MyKey InstanceType=t1.micro"

    # Regex that parses CloudFormation parameter key-value pairs:
    # https://regex101.com/r/xqfSjW/2
    # https://regex101.com/r/xqfSjW/5

    # If Both ParameterKey pattern and KeyPairName=MyKey should not be present
    # while adding parameter overrides, if they are, it
    # can result in unpredicatable behavior.
    # Use this regex when you have space as delimiter Ex: "KeyName1=string KeyName2=string"
    VALUE_REGEX_SPACE_DELIM = _generate_match_regex(match_pattern=".", delim=" ")
    _pattern_1 = r"(?:ParameterKey={key},ParameterValue={value})".format(
        key=PARAM_AND_METADATA_KEY_REGEX, value=VALUE_REGEX_SPACE_DELIM
    )
    _pattern_2 = r"(?:(?: ){key}={value})".format(key=PARAM_AND_METADATA_KEY_REGEX, value=VALUE_REGEX_SPACE_DELIM)

    ordered_pattern_match = [_pattern_1, _pattern_2]

    # NOTE(TheSriram): name needs to be added to click.ParamType requires it.
    name = ""

    def convert(self, value, param, ctx):
        result = {}

        # Empty tuple
        if value == ("",):
            return result

        value = (value,) if isinstance(value, str) else value
        for val in value:
            val.strip()
            # Add empty string to start of the string to help match `_pattern2`
            val = " " + val

            try:
                # NOTE(TheSriram): find the first regex that matched.
                # pylint is concerned that we are checking at the same `val` within the loop,
                # but that is the point, so disabling it.
                pattern = next(
                    i
                    for i in filter(
                        lambda item: re.findall(item, val), self.ordered_pattern_match
                    )  # pylint: disable=cell-var-from-loop
                )
            except StopIteration:
                return self.fail(
                    "{} is not in valid format. It must look something like '{}' or '{}'".format(
                        val, self.__EXAMPLE_1, self.__EXAMPLE_2
                    ),
                    param,
                    ctx,
                )

            groups = re.findall(pattern, val)

            # 'groups' variable is a list of tuples ex: [(key1, value1), (key2, value2)]
            for key, param_value in groups:
                result[_unquote_wrapped_quotes(key)] = _unquote_wrapped_quotes(param_value)

        return result


class CfnMetadataType(click.ParamType):
    """
    Custom Click options type to accept values for metadata parameters.
    metadata parameters can be of the type KeyName1=string,KeyName2=string or {"string":"string"}
    """

    _EXAMPLE = 'KeyName1=string,KeyName2=string or {"string":"string"}'
    # Use this regex when you have comma as delimiter Ex: "KeyName1=string,KeyName2=string"
    VALUE_REGEX_COMMA_DELIM = _generate_match_regex(match_pattern=".", delim=",")

    _pattern = r"(?:{key}={value})".format(key=PARAM_AND_METADATA_KEY_REGEX, value=VALUE_REGEX_COMMA_DELIM)

    # NOTE(TheSriram): name needs to be added to click.ParamType requires it.
    name = ""

    def convert(self, value, param, ctx):
        result = {}
        fail = False
        if not value:
            return result
        try:
            # Look to load the value into json if we can.
            result = json.loads(value)
            for val in result.values():
                if isinstance(val, (dict, list)):
                    # Need a non nested dictionary or a dictionary with non list values,
                    # If either is found, fail the conversion.
                    fail = True
        except JSONDecodeError:
            # if looking for a json format failed, look at if the specified value follows
            # KeyName1=string,KeyName2=string format
            groups = re.findall(self._pattern, value)

            if not groups:
                fail = True
            for group in groups:
                key, v = group
                # assign to result['KeyName1'] = string and so on.
                result[key] = v

        if fail:
            return self.fail(
                "{} is not in valid format. It must look something like '{}'".format(value, self._EXAMPLE), param, ctx
            )

        return result


class CfnTags(click.ParamType):
    """
    Custom Click options type to accept values for tag parameters.
    tag parameters can be of the type KeyName1=string KeyName2=string

    If multiple_values_per_key is set to True, the returned dictionary will map
    each key to a list of corresponding values.

    E.g. Input: KeyName1=Value1 KeyName1=Value2 Output: {KeyName1 : [Value1, Value2]}
    """

    def __init__(self, multiple_values_per_key=False):
        self.multiple_values_per_key = multiple_values_per_key

    _EXAMPLE = "KeyName1=string KeyName2=string"
    # Tags have additional constraints and they allow "+ - = . _ : / @" apart from alpha-numerics.
    TAG_REGEX = '[A-Za-z0-9\\"_:\\.\\/\\+-\\@=]'

    _pattern = r"{tag}={tag}".format(tag=_generate_match_regex(match_pattern=TAG_REGEX, delim=" "))

    # NOTE(TheSriram): name needs to be added to click.ParamType requires it.
    name = ""

    def convert(self, value, param, ctx):
        result = {}
        fail = False
        # Empty tuple
        if value == ("",):
            return result

        # if value comes in via samconfig file and is a list, it is parsed to string.
        if isinstance(value, list):
            if not value:
                return result
            value = " ".join(value)
        # if value comes in a via configuration file, it will be a string. So we should still convert it.
        value = (value,) if not isinstance(value, tuple) else value

        for val in value:
            # Using standard parser first.
            # We should implement other type parser like JSON and Key=key,Value=val type format.
            parsed, tags = self._standard_key_value_parser(val)
            if not parsed:
                parsed, tags = self._space_separated_key_value_parser(val)
            if parsed:
                for k in tags:
                    self._add_value(result, _unquote_wrapped_quotes(k), _unquote_wrapped_quotes(tags[k]))
            else:
                groups = re.findall(self._pattern, val)

                if not groups:
                    fail = True
                for group in groups:
                    key, v = group
                    self._add_value(result, _unquote_wrapped_quotes(key), _unquote_wrapped_quotes(v))

            if fail:
                return self.fail(
                    "{} is not in valid format. It must look something like '{}'".format(value, self._EXAMPLE),
                    param,
                    ctx,
                )

        return result

    def _add_value(self, result: dict, key: str, new_value: str):
        """
        Add a given value to a given key in the result map.
        """
        if self.multiple_values_per_key:
            if not result.get(key):
                result[key] = []
            result[key].append(new_value)
            return
        result[key] = new_value

    @staticmethod
    def _standard_key_value_parser(tag_value):
        """
        Method to parse simple `Key=Value` type tags without using regex. This is similar to how aws-cli does this.
        https://github.com/aws/aws-cli/blob/eff79a263347e8e83c8a2cc07265ab366315a992/awscli/customizations/cloudformation/deploy.py#L361
        Parameters
        ----------
        tag_value

        Returns
        -------

        """
        equals_count = tag_value.count("=")
        if equals_count != 1:
            return False, None

        splits = tag_value.split("=")
        return True, {splits[0]: splits[1]}

    @staticmethod
    def _space_separated_key_value_parser(tag_value):
        """
        Method to parse space separated `Key1=Value1 Key2=Value2` type tags without using regex.
        Parameters
        ----------
        tag_value
        """
        tags_dict = {}
        for value in tag_value.split(" "):
            parsed, parsed_tag = CfnTags._standard_key_value_parser(value)
            if not parsed:
                return False, None
            tags_dict = {**tags_dict, **parsed_tag}
        return True, tags_dict


class SigningProfilesOptionType(click.ParamType):
    """
    Custom parameter type to parse Signing Profile options
    Example options could be;
    - MyFunctionOrLayerToBeSigned=MySigningProfile
    - MyFunctionOrLayerToBeSigned=MySigningProfile:MyProfileOwner

    See convert function docs for details
    """

    pattern = r"(?:(?: )([A-Za-z0-9\"]+)=(\"(?:\\.|[^\"\\]+)*\"|(?:\\.|[^ \"\\]+)+))"

    # Note: this is required, otherwise it is failing when running help
    name = ""

    def convert(self, value, param, ctx):
        """
        Converts given Signing Profile options to a dictionary where Function or Layer name would be key,
        and signing profile details would be the value.

        Since this method is also been used when we are reading the config details from samconfig.toml,
        If value is already a dictionary, don't need to do anything, just return it.
        If value is an array, then each value will correspond to a function or layer config, and here it is parsed
        and converted into a dictionary
        """
        result = {}

        # Empty tuple
        if value == ("",):
            return result

        value = (value,) if isinstance(value, str) else value
        for val in value:
            val.strip()
            # Add empty string to start of the string to help match `_pattern2`
            val = " " + val

            signing_profiles = re.findall(self.pattern, val)

            # if no signing profiles found by regex, then fail
            if not signing_profiles:
                return self.fail(
                    f"{value} is not a valid code sign config, it should look like this 'MyFunction=SigningProfile'",
                    param,
                    ctx,
                )

            for function_name, param_value in signing_profiles:
                (signer_profile_name, signer_profile_owner) = self._split_signer_profile_name_owner(
                    _unquote_wrapped_quotes(param_value)
                )

                # code signing requires profile name, if it is not present then fail
                if not signer_profile_name:
                    return self.fail(
                        "Signer profile option has invalid format, it should look like this "
                        "MyFunction=MySigningProfile or MyFunction=MySigningProfile:MySigningProfileOwner",
                        param,
                        ctx,
                    )

                result[_unquote_wrapped_quotes(function_name)] = {
                    "profile_name": signer_profile_name,
                    "profile_owner": signer_profile_owner,
                }

        return result

    @staticmethod
    def _split_signer_profile_name_owner(signing_profile):
        equals_count = signing_profile.count(":")

        if equals_count > 1:
            return None, None
        if equals_count == 1:
            return signing_profile.split(":")
        return signing_profile, ""


class ImageRepositoryType(click.ParamType):
    """
    Custom Parameter Type for Image Repository option.
    """

    class Transformer:
        """
        Class takes in a converter (native click type) and a transformer.
        transformer is a function callback where additional transformation happens before
        conversion to a native click type.
        """

        def __init__(self, converter, transformation):
            """

            :param converter: native click Type
            :param transformation: callback function for transformation prior to conversion.
            """
            self.converter = converter
            self.transformer = transformation

        def transform(self, *args, **kwargs):
            return self.transformer(self.converter.convert(*args, **kwargs))

    # Transformation callback function checks if the received option value is a valid ECR url.
    transformer = Transformer(converter=click.STRING, transformation=is_ecr_url)
    name = ""

    def convert(self, value, param, ctx):
        """
        Attempt a conversion given the stipulations of allowed transformations.
        """
        result = self.transformer.transform(value, param, ctx)
        if not result:
            raise click.BadParameter(f"Invalid Image Repository ECR URI: {value}")
        return value


class ImageRepositoriesType(click.ParamType):
    """
    Custom Parameter Type for Multi valued Image Repositories option.
    """

    name = ""
    KEY_VALUE_PAIR_LENGTH = 2

    def convert(self, value, param, ctx):
        key_value_pair = value.split("=")
        if len(key_value_pair) != self.KEY_VALUE_PAIR_LENGTH:
            raise click.BadParameter(
                f"{param.opts[0]} is not a valid format, it needs to be of the form function_logical_id=ECR_URI"
            )
        key = key_value_pair[0]
        _value = key_value_pair[1]
        if not is_ecr_url(_value):
            raise click.BadParameter(f"{param.opts[0]} needs to have valid ECR URI as value")
        return {key: _value}



"""
CLI command for "logs" command
"""

import logging

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, pass_context, print_cmdline_args
from samcli.cli.main import common_options as cli_framework_options
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands._utils.options import common_observability_options, generate_next_command_recommendation
from samcli.commands.logs.validation_and_exception_handlers import (
    SAM_LOGS_ADDITIONAL_EXCEPTION_HANDLERS,
    stack_name_cw_log_group_validation,
)
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version

LOG = logging.getLogger(__name__)

HELP_TEXT = """
Use this command to fetch logs generated by your Lambda function.\n
\b
When your functions are a part of a CloudFormation stack, you can fetch logs using the function's
LogicalID when you specify the stack name.
$ sam logs -n HelloWorldFunction --stack-name mystack \n
\b
Or, you can fetch logs using the function's name.
$ sam logs -n mystack-HelloWorldFunction-1FJ8PD36GML2Q \n
\b
You can view logs for a specific time range using the -s (--start-time) and -e (--end-time) options
$ sam logs -n HelloWorldFunction --stack-name mystack -s '10min ago' -e '2min ago' \n
\b
You can also add the --tail option to wait for new logs and see them as they arrive.
$ sam logs -n HelloWorldFunction --stack-name mystack --tail \n
\b
Use the --filter option to quickly find logs that match terms, phrases or values in your log events.
$ sam logs -n HelloWorldFunction --stack-name mystack --filter 'error' \n
\b
Fetch logs for all supported resources in your application, and additionally from the specified log groups.
$ sam logs --cw-log-group /aws/lambda/myfunction-123 --cw-log-group /aws/lambda/myfunction-456
\b
You can now fetch logs from supported resources, by only providing --stack-name parameter
$ sam logs --stack-name mystack \n
\b
You can also fetch logs from a resource which is defined in a nested stack.
$ sam logs --stack-name mystack -n MyNestedStack/HelloWorldFunction
"""


@click.command("logs", help=HELP_TEXT, short_help="Fetch logs for a function")
@configuration_option(provider=TomlProvider(section="parameters"))
@click.option(
    "--name",
    "-n",
    multiple=True,
    help="The name of the resource for which to fetch logs. If this resource is a part of an AWS CloudFormation stack, "
    "this can be the LogicalID of the resource in the CloudFormation/SAM template. "
    "Multiple names can be provided by repeating the parameter again. "
    "If resource is in a nested stack, name can be prepended by nested stack name to pull logs "
    "from that resource (NestedStackLogicalId/ResourceLogicalId). "
    "If it is not provided and no --cw-log-group have been given, it will scan "
    "given stack and find all supported resources, and start pulling log information from them.",
)
@click.option("--stack-name", default=None, help="Name of the AWS CloudFormation stack that the function is a part of.")
@click.option(
    "--filter",
    default=None,
    help="You can specify an expression to quickly find logs that match terms, phrases or values in "
    'your log events. This could be a simple keyword (e.g. "error") or a pattern '
    "supported by AWS CloudWatch Logs. See the AWS CloudWatch Logs documentation for the syntax "
    "https://docs.aws.amazon.com/AmazonCloudWatch/latest/logs/FilterAndPatternSyntax.html",
)
@click.option(
    "--include-traces",
    "-i",
    is_flag=True,
    help="Include the XRay traces in the log output.",
)
@click.option(
    "--cw-log-group",
    multiple=True,
    help="Additional CloudWatch Log group names that are not auto-discovered based upon --name parameter. "
    "When provided, it will only tail the given CloudWatch Log groups. If you want to tail log groups related "
    "to resources, please also provide their names as well",
)
@common_observability_options
@cli_framework_options
@aws_creds_options
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
@command_exception_handler(SAM_LOGS_ADDITIONAL_EXCEPTION_HANDLERS)
@stack_name_cw_log_group_validation
def cli(
    ctx,
    name,
    stack_name,
    filter,
    tail,
    include_traces,
    start_time,
    end_time,
    output,
    cw_log_group,
    config_file,
    config_env,
):  # pylint: disable=redefined-builtin
    """
    `sam logs` command entry point
    """
    # All logic must be implemented in the ``do_cli`` method. This helps with easy unit testing

    do_cli(
        name,
        stack_name,
        filter,
        tail,
        include_traces,
        start_time,
        end_time,
        cw_log_group,
        output,
        ctx.region,
        ctx.profile,
    )  # pragma: no cover


def do_cli(
    names,
    stack_name,
    filter_pattern,
    tailing,
    include_tracing,
    start_time,
    end_time,
    cw_log_groups,
    output,
    region,
    profile,
):
    """
    Implementation of the ``cli`` method
    """

    from datetime import datetime

    from samcli.commands.logs.logs_context import ResourcePhysicalIdResolver, parse_time
    from samcli.commands.logs.puller_factory import generate_puller
    from samcli.lib.observability.util import OutputOption
    from samcli.lib.utils.boto_utils import get_boto_client_provider_with_config, get_boto_resource_provider_with_config

    if names and len(names) <= 1:
        click.echo(
            "You can now use 'sam logs' without --name parameter, "
            "which will pull the logs from all supported resources in your stack."
        )

    sanitized_start_time = parse_time(start_time, "start-time")
    sanitized_end_time = parse_time(end_time, "end-time") or datetime.utcnow()

    boto_client_provider = get_boto_client_provider_with_config(region=region, profile=profile)
    boto_resource_provider = get_boto_resource_provider_with_config(region=region, profile=profile)
    resource_logical_id_resolver = ResourcePhysicalIdResolver(
        boto_resource_provider, boto_client_provider, stack_name, names
    )

    # only fetch all resources when no CloudWatch log group defined
    fetch_all_when_no_resource_name_given = not cw_log_groups
    puller = generate_puller(
        boto_client_provider,
        resource_logical_id_resolver.get_resource_information(fetch_all_when_no_resource_name_given),
        filter_pattern,
        cw_log_groups,
        OutputOption(output) if output else OutputOption.text,
        include_tracing,
    )

    if tailing:
        puller.tail(sanitized_start_time, filter_pattern)
    else:
        puller.load_time_period(sanitized_start_time, sanitized_end_time, filter_pattern)

    if tailing:
        command_suggestions = generate_next_command_recommendation(
            [
                (
                    "Tail Logs from All Support Resources and X-Ray",
                    f"sam logs --stack-name {stack_name} --tail --include-traces",
                ),
                ("Tail X-Ray Information", "sam traces --tail"),
            ]
        )
        click.secho(command_suggestions, fg="yellow")



"""
Factory methods which generates puller and consumer instances for XRay events
"""
from typing import Any, List

from samcli.commands.traces.trace_console_consumers import XRayTraceConsoleConsumer
from samcli.lib.observability.observability_info_puller import (
    ObservabilityCombinedPuller,
    ObservabilityEventConsumer,
    ObservabilityEventConsumerDecorator,
    ObservabilityPuller,
)
from samcli.lib.observability.util import OutputOption
from samcli.lib.observability.xray_traces.xray_event_mappers import (
    XRayServiceGraphConsoleMapper,
    XRayServiceGraphJSONMapper,
    XRayTraceConsoleMapper,
    XRayTraceJSONMapper,
)
from samcli.lib.observability.xray_traces.xray_event_puller import XRayTracePuller
from samcli.lib.observability.xray_traces.xray_service_graph_event_puller import XRayServiceGraphPuller


def generate_trace_puller(
    xray_client: Any,
    output: OutputOption = OutputOption.text,
) -> ObservabilityPuller:
    """
    Generates puller instance with correct consumer and/or mapper configuration

    Parameters
    ----------
    xray_client : Any
        boto3 xray client to be used in XRayTracePuller instance
    output : OutputOption
        Decides how the output will be presented in the console. It is been used to select correct consumer type
        between (default) text consumer or json consumer

    Returns
    -------
        Puller instance with desired configuration
    """
    pullers: List[ObservabilityPuller] = []
    pullers.append(XRayTracePuller(xray_client, generate_xray_event_consumer(output)))
    pullers.append(XRayServiceGraphPuller(xray_client, generate_xray_service_graph_consumer(output)))

    return ObservabilityCombinedPuller(pullers)


def generate_json_xray_event_consumer() -> ObservabilityEventConsumer:
    """
    Generates unformatted consumer, which will print XRay events unformatted JSON into terminal

    Returns
    -------
        File consumer instance with desired mapper configuration
    """
    return ObservabilityEventConsumerDecorator([XRayTraceJSONMapper()], XRayTraceConsoleConsumer())


def generate_xray_event_console_consumer() -> ObservabilityEventConsumer:
    """
    Generates an instance of event consumer which will print events into console

    Returns
    -------
        Console consumer instance with desired mapper configuration
    """
    return ObservabilityEventConsumerDecorator([XRayTraceConsoleMapper()], XRayTraceConsoleConsumer())


def generate_xray_event_consumer(output: OutputOption = OutputOption.text) -> ObservabilityEventConsumer:
    """
    Generates consumer instance with the given variables.
    If output is JSON, then it will return consumer with formatters for just JSON.
    Otherwise, it will return regular text console consumer
    """
    if output == OutputOption.json:
        return generate_json_xray_event_consumer()
    return generate_xray_event_console_consumer()


def generate_json_xray_service_graph_consumer() -> ObservabilityEventConsumer:
    """
    Generates unformatted consumer, which will print XRay events unformatted JSON into terminal

    Returns
    -------
        File consumer instance with desired mapper configuration
    """
    return ObservabilityEventConsumerDecorator([XRayServiceGraphJSONMapper()], XRayTraceConsoleConsumer())


def generate_xray_service_graph_console_consumer() -> ObservabilityEventConsumer:
    """
    Generates an instance of event consumer which will print events into console

    Returns
    -------
        Console consumer instance with desired mapper configuration
    """
    return ObservabilityEventConsumerDecorator([XRayServiceGraphConsoleMapper()], XRayTraceConsoleConsumer())


def generate_xray_service_graph_consumer(output: OutputOption = OutputOption.text) -> ObservabilityEventConsumer:
    """
    Generates consumer instance with the given variables.
    If output is JSON, then it will return consumer with formatters for just JSON.
    Otherwise, it will return regular text console consumer
    """
    if output == OutputOption.json:
        return generate_json_xray_service_graph_consumer()
    return generate_xray_service_graph_console_consumer()



"""
CloudWatch log event type
"""
from typing import Optional

from samcli.lib.observability.observability_info_puller import ObservabilityEvent


class CWLogEvent(ObservabilityEvent[dict]):
    """
    An event class which represents a Cloud Watch log
    """

    def __init__(self, cw_log_group: str, event: dict, resource_name: Optional[str] = None):
        """
        Parameters
        ----------
        cw_log_group : str
            Name of the CloudWatch log group
        event : dict
            Event dictionary of the CloudWatch log event
        resource_name : Optional[str]
            Resource name that is related to this CloudWatch log event
        """
        self.cw_log_group = cw_log_group
        self.message: str = event.get("message", "")
        self.log_stream_name: str = event.get("logStreamName", "")
        timestamp: int = event.get("timestamp", 0)
        super().__init__(event, timestamp, resource_name)

    def __eq__(self, other):
        if not isinstance(other, CWLogEvent):
            return False

        return (
            self.cw_log_group == other.cw_log_group
            and self.log_stream_name == other.log_stream_name
            and self.timestamp == other.timestamp
            and self.message == other.message
        )



"""SyncFlow for Image based Lambda Functions"""
import logging
from contextlib import ExitStack
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import docker
from docker.client import DockerClient

from samcli.lib.build.app_builder import ApplicationBuilder
from samcli.lib.package.ecr_uploader import ECRUploader
from samcli.lib.providers.provider import Stack
from samcli.lib.sync.flows.function_sync_flow import FunctionSyncFlow, wait_for_function_update_complete
from samcli.lib.sync.sync_flow import ApiCallTypes, ResourceAPICall

if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.sync.sync_context import SyncContext

LOG = logging.getLogger(__name__)


class ImageFunctionSyncFlow(FunctionSyncFlow):
    _ecr_client: Optional[Any]
    _docker_client: Optional[DockerClient]
    _image_name: Optional[str]

    def __init__(
        self,
        function_identifier: str,
        build_context: "BuildContext",
        deploy_context: "DeployContext",
        sync_context: "SyncContext",
        physical_id_mapping: Dict[str, str],
        stacks: List[Stack],
    ):
        """
        Parameters
        ----------
        function_identifier : str
            Image function resource identifier that need to be synced.
        build_context : BuildContext
            BuildContext
        deploy_context : DeployContext
            DeployContext
        sync_context: SyncContext
            SyncContext object that obtains sync information.
        physical_id_mapping : Dict[str, str]
            Physical ID Mapping
        stacks : Optional[List[Stack]]
            Stacks
        """
        super().__init__(function_identifier, build_context, deploy_context, sync_context, physical_id_mapping, stacks)
        self._ecr_client = None
        self._image_name = None
        self._docker_client = None

    def _get_docker_client(self) -> DockerClient:
        """Lazy instantiates and returns the docker client"""
        if not self._docker_client:
            self._docker_client = docker.from_env()
        return self._docker_client

    def _get_ecr_client(self) -> Any:
        """Lazy instantiates and returns the boto3 ecr client"""
        if not self._ecr_client:
            self._ecr_client = self._boto_client("ecr")
        return self._ecr_client

    def gather_resources(self) -> None:
        """Build function image and save it in self._image_name"""
        builder = ApplicationBuilder(
            self._build_context.collect_build_resources(self._function_identifier),
            self._build_context.build_dir,
            self._build_context.base_dir,
            self._build_context.cache_dir,
            cached=False,
            is_building_specific_resource=True,
            manifest_path_override=self._build_context.manifest_path_override,
            container_manager=self._build_context.container_manager,
            mode=self._build_context.mode,
            build_in_source=self._build_context.build_in_source,
        )
        self._image_name = builder.build().artifacts.get(self._function_identifier)
        if self._image_name:
            self._local_sha = self._get_local_image_id(self._image_name)

    def _get_local_image_id(self, image: str) -> Optional[str]:
        """Returns the local hash of the image"""
        docker_img = self._get_docker_client().images.get(image)
        if not docker_img or not docker_img.attrs.get("Id"):
            return None
        return str(docker_img.attrs.get("Id"))

    def compare_remote(self) -> bool:
        return False

    def sync(self) -> None:
        if not self._image_name:
            LOG.debug("%sSkipping sync. Image name is None.", self.log_prefix)
            return
        function_physical_id = self.get_physical_id(self._function_identifier)
        # Load ECR Repo from --image-repository
        ecr_repo = self._deploy_context.image_repository

        # Load ECR Repo from --image-repositories
        if (
            not ecr_repo
            and self._deploy_context.image_repositories
            and isinstance(self._deploy_context.image_repositories, dict)
        ):
            ecr_repo = self._deploy_context.image_repositories.get(self._function_identifier)

        # Load ECR Repo directly from remote function
        if not ecr_repo:
            LOG.debug("%sGetting ECR Repo from Remote Function", self.log_prefix)
            function_result = self._lambda_client.get_function(FunctionName=function_physical_id)
            ecr_repo = function_result.get("Code", dict()).get("ImageUri", "").split(":")[0]
        ecr_uploader = ECRUploader(self._get_docker_client(), self._get_ecr_client(), ecr_repo, None)
        image_uri = ecr_uploader.upload(self._image_name, self._function_identifier)

        with ExitStack() as exit_stack:
            if self.has_locks():
                exit_stack.enter_context(self._get_lock_chain())

            self._lambda_client.update_function_code(FunctionName=function_physical_id, ImageUri=image_uri)

            # We need to wait for the cloud side update to finish
            # Otherwise even if the call is finished and lockchain is released
            # It is still possible that we have a race condition on cloud updating the same function
            wait_for_function_update_complete(self._lambda_client, self.get_physical_id(self._function_identifier))

    def _get_resource_api_calls(self) -> List[ResourceAPICall]:
        # We need to acquire lock for both API calls since they would conflict on cloud
        # Any UPDATE_FUNCTION_CODE and UPDATE_FUNCTION_CONFIGURATION on the same function
        # Cannot take place in parallel
        return [
            ResourceAPICall(
                self._function_identifier,
                [ApiCallTypes.UPDATE_FUNCTION_CODE, ApiCallTypes.UPDATE_FUNCTION_CONFIGURATION],
            )
        ]



"""
StackBuilder implementation for nested stack
"""
import re
from typing import cast

from samcli.lib.bootstrap.stack_builder import AbstractStackBuilder
from samcli.lib.providers.provider import Function
from samcli.lib.utils.hash import str_checksum
from samcli.lib.utils.resources import AWS_CLOUDFORMATION_STACK, AWS_SERVERLESS_LAYERVERSION

CREATED_BY_METADATA_KEY = "CreatedBy"
CREATED_BY_METADATA_VALUE = "AWS SAM CLI sync command"
NON_ALPHANUM_REGEX = re.compile(r"\W+")


class NestedStackBuilder(AbstractStackBuilder):
    """
    CFN/SAM Template creator for nested stack
    """

    def __init__(self):
        super().__init__("AWS SAM CLI Nested Stack for Auto Dependency Layer Creation")
        self.add_metadata(CREATED_BY_METADATA_KEY, CREATED_BY_METADATA_VALUE)

    def is_any_function_added(self) -> bool:
        return bool(self._template_dict.get("Resources", {}))

    def add_function(
        self,
        stack_name: str,
        layer_contents_folder: str,
        function: Function,
    ) -> str:
        layer_logical_id = self.get_layer_logical_id(function.full_path)
        layer_name = self.get_layer_name(stack_name, function.full_path)

        self.add_resource(
            layer_logical_id,
            self._get_layer_dict(function.full_path, layer_name, layer_contents_folder, cast(str, function.runtime)),
        )
        self.add_output(layer_logical_id, {"Ref": layer_logical_id})
        return layer_logical_id

    @staticmethod
    def get_layer_logical_id(function_logical_id: str) -> str:
        function_logical_id_hash = str_checksum(function_logical_id)
        sanitized_function_logical_id = NestedStackBuilder._get_logical_id_compliant_str(function_logical_id)
        return f"{sanitized_function_logical_id[:48]}{function_logical_id_hash[:8]}DepLayer"

    @staticmethod
    def get_layer_name(stack_name: str, function_logical_id: str) -> str:
        function_logical_id_hash = str_checksum(function_logical_id)
        sanitized_function_logical_id = NestedStackBuilder._get_logical_id_compliant_str(function_logical_id)
        stack_name_hash = str_checksum(stack_name)
        return (
            f"{stack_name[:16]}{stack_name_hash[:8]}-{sanitized_function_logical_id[:22]}{function_logical_id_hash[:8]}"
            f"-DepLayer"
        )

    @staticmethod
    def _get_layer_dict(function_logical_id: str, layer_name: str, layer_contents_folder: str, function_runtime: str):
        return {
            "Type": AWS_SERVERLESS_LAYERVERSION,
            "Properties": {
                "LayerName": layer_name,
                "Description": f"Auto created layer for dependencies of function {function_logical_id}",
                "ContentUri": layer_contents_folder,
                "RetentionPolicy": "Delete",
                "CompatibleRuntimes": [function_runtime],
            },
            "Metadata": {CREATED_BY_METADATA_KEY: CREATED_BY_METADATA_VALUE},
        }

    @staticmethod
    def get_nested_stack_reference_resource(nested_template_location):
        return {
            "Type": AWS_CLOUDFORMATION_STACK,
            "DeletionPolicy": "Delete",
            "Properties": {"TemplateURL": nested_template_location},
            "Metadata": {CREATED_BY_METADATA_KEY: CREATED_BY_METADATA_VALUE},
        }

    @staticmethod
    def _get_logical_id_compliant_str(function_logical_id: str):
        """
        Removes all non-alphanumeric chars to make it usable for resource name definition
        https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/resources-section-structure.html
        """
        return NON_ALPHANUM_REGEX.sub("", function_logical_id)



"""
Helper methods that aid interactions within docker containers.
"""

import logging
import os
import pathlib
import platform
import posixpath
import random
import re
import socket

import docker
import requests

from samcli.lib.utils.architecture import ARM64, validate_architecture
from samcli.local.docker.exceptions import NoFreePortsError

LOG = logging.getLogger(__name__)


def to_posix_path(code_path):
    """
    Change the code_path to be of unix-style if running on windows when supplied with an absolute windows path.

    Parameters
    ----------
    code_path : str
        Directory in the host operating system that should be mounted within the container.
    Returns
    -------
    str
        Posix equivalent of absolute windows style path.
    Examples
    --------
    >>> to_posix_path('/Users/UserName/sam-app')
    /Users/UserName/sam-app
    >>> to_posix_path('C:\\\\Users\\\\UserName\\\\AppData\\\\Local\\\\Temp\\\\mydir')
    /c/Users/UserName/AppData/Local/Temp/mydir
    """

    return (
        re.sub(
            "^([A-Za-z])+:",
            lambda match: posixpath.sep + match.group().replace(":", "").lower(),
            pathlib.PureWindowsPath(code_path).as_posix(),
        )
        if os.name == "nt"
        else code_path
    )


def find_free_port(start=5000, end=9000):
    """
    Utility function which scans through a port range in a randomized manner
    and finds the first free port a socket can bind to.
    :raises NoFreePortException if no free ports found in range.
    :return: int - free port
    """
    port_range = [random.randrange(start, end) for _ in range(start, end)]
    for port in port_range:
        try:
            s = socket.socket()
            s.bind(("", port))
            s.close()
            return port
        except OSError:
            continue
    raise NoFreePortsError(f"No free ports on the host machine from {start} to {end}")


def is_docker_reachable(docker_client):
    """
    Checks if Docker daemon is running.

    :param docker_client : docker.from_env() - docker client object
    :returns True, if Docker is available, False otherwise.
    """
    errors = (
        docker.errors.APIError,
        requests.exceptions.ConnectionError,
    )
    if platform.system() == "Windows":
        import pywintypes  # pylint: disable=import-error

        errors += (pywintypes.error,)

    try:
        docker_client.ping()
        return True

    # When Docker is not installed, a request.exceptions.ConnectionError is thrown.
    # and also windows-specific errors
    except errors:
        LOG.debug("Docker is not reachable", exc_info=True)
        return False


def get_rapid_name(architecture: str) -> str:
    """
    Return the name of the rapid binary to use for an architecture

    Parameters
    ----------
    architecture : str
        Architecture

    Returns
    -------
    str
        "aws-lambda-rie-" + architecture
    """
    validate_architecture(architecture)

    return "aws-lambda-rie-" + architecture


def get_image_arch(architecture: str) -> str:
    """
    Returns the docker image architecture value corresponding to the
    Lambda architecture value

    Parameters
    ----------
    architecture : str
        Lambda architecture

    Returns
    -------
    str
        Docker image architecture
    """
    validate_architecture(architecture)

    return "arm64" if architecture == ARM64 else "amd64"


def get_docker_platform(architecture: str) -> str:
    """
    Returns the platform to pass to the docker client for a given architecture

    Parameters
    ----------
    architecture : str
        Architecture

    Returns
    -------
    str
        linux/arm64 for arm64, linux/amd64 otherwise
    """
    validate_architecture(architecture)

    return f"linux/{get_image_arch(architecture)}"



"""
Read Swagger documents from variety of sources
"""

import logging
import os
import tempfile
from urllib.parse import parse_qs, urlparse

import boto3
import botocore

from samcli.yamlhelper import yaml_parse

LOG = logging.getLogger(__name__)
_FN_TRANSFORM = "Fn::Transform"


def parse_aws_include_transform(data):
    """
    If the input data is an AWS::Include data, then parse and return the location of the included file.

    AWS::Include transform data usually has the following format:
    {
        "Fn::Transform": {
            "Name": "AWS::Include",
            "Parameters": {
                "Location": "s3://MyAmazonS3BucketName/swagger.yaml"
            }
        }
    }

    Parameters
    ----------
    data : dict
        Dictionary data to parse

    Returns
    -------
    str
        Location of the included file, if available. None, otherwise
    """

    if not data:
        return None

    if _FN_TRANSFORM not in data:
        return None

    transform_data = data[_FN_TRANSFORM]

    name = transform_data.get("Name")
    location = transform_data.get("Parameters", {}).get("Location")
    if name == "AWS::Include":
        LOG.debug("Successfully parsed location from AWS::Include transform: %s", location)
        return location

    return None


class SwaggerReader:
    """
    Class to read and parse Swagger document from a variety of sources. This class accepts the same data formats as
    available in Serverless::Api SAM resource
    """

    def __init__(self, definition_body=None, definition_uri=None, working_dir=None):
        """
        Initialize the class with swagger location

        Parameters
        ----------
        definition_body : dict
            Swagger document as a dictionary directly or inlined using AWS::Include transform.

        definition_uri : str or dict
            Location of the Swagger file. Supports three formats:
                - S3 URI Ex: ``s3://mybucket/swagger.yaml``
                - S3 URI as a dictionary Ex: ``{"Bucket": "mybucket", "Key": "swagger.yaml", "Version": "123"}``
                - Local file path either as absolute or relative path Ex: ``./swagger.yaml``. Relative paths are
                  resolved to with respect to the given working directory.

        working_dir : str
            Path to the working directory with respect which we will resolve local relative paths
        """
        self.definition_body = definition_body
        self.definition_uri = definition_uri
        self.working_dir = working_dir

        if not self.definition_body and not self.definition_uri:
            raise ValueError("Require value for either DefinitionBody or DefinitionUri")

    def read(self):
        """
        Gets the Swagger document from either of the given locations. If we fail to retrieve or parse the Swagger
        file, this method will return None.

        Returns
        -------
        dict:
            Swagger document. None, if we cannot retrieve the document
        """

        swagger = None

        # First check if there is inline swagger
        if self.definition_body:
            swagger = self._read_from_definition_body()

        if not swagger and self.definition_uri:
            # If not, then try to download it from the given URI
            swagger = self._download_swagger(self.definition_uri)

        return swagger

    def _read_from_definition_body(self):
        """
        Read the Swagger document from DefinitionBody. It could either be an inline Swagger dictionary or an
        AWS::Include macro that contains location of the included Swagger. In the later case, we will download and
        parse the Swagger document.

        Returns
        -------
        dict
            Swagger document, if we were able to parse. None, otherwise
        """

        # Let's try to parse it as AWS::Include Transform first. If not, then fall back to assuming the Swagger document
        # was inclined directly into the body
        location = parse_aws_include_transform(self.definition_body)
        if location:
            LOG.debug("Trying to download Swagger from %s", location)
            return self._download_swagger(location)

        # Inline Swagger, just return the contents which should already be a dictionary
        LOG.debug("Detected Inline Swagger definition")
        return self.definition_body

    def _download_swagger(self, location):
        """
        Download the file from given local or remote location and return it

        Parameters
        ----------
        location : str or dict
            Local path or S3 path to Swagger file to download. Consult the ``__init__.py`` documentation for specifics
            on structure of this property.

        Returns
        -------
        dict or None
            Downloaded and parsed Swagger document. None, if unable to download
        """

        if not location:
            return None

        bucket, key, version = self._parse_s3_location(location)
        if bucket and key:
            LOG.debug("Downloading Swagger document from Bucket=%s, Key=%s, Version=%s", bucket, key, version)
            swagger_str = self._download_from_s3(bucket, key, version)
            return yaml_parse(swagger_str)

        if not isinstance(location, str):
            # This is not a string and not a S3 Location dictionary. Probably something invalid
            LOG.debug("Unable to download Swagger file. Invalid location: %s", location)
            return None

        # ``location`` is a string and not a S3 path. It is probably a local path. Let's resolve relative path if any
        filepath = location
        if self.working_dir:
            # Resolve relative paths, if any, with respect to working directory
            filepath = os.path.join(self.working_dir, location)

        if not os.path.exists(filepath):
            LOG.debug("Unable to download Swagger file. File not found at location %s", filepath)
            return None

        LOG.debug("Reading Swagger document from local file at %s", filepath)
        with open(filepath, "r") as fp:
            return yaml_parse(fp.read())

    @staticmethod
    def _download_from_s3(bucket, key, version=None):
        """
        Download a file from given S3 location, if available.

        Parameters
        ----------
        bucket : str
            S3 Bucket name

        key : str
            S3 Bucket Key aka file path

        version : str
            Optional Version ID of the file

        Returns
        -------
        str
            Contents of the file that was downloaded

        Raises
        ------
        botocore.exceptions.ClientError if we were unable to download the file from S3
        """

        s3 = boto3.client("s3")

        extra_args = {}
        if version:
            extra_args["VersionId"] = version

        with tempfile.TemporaryFile() as fp:
            try:
                s3.download_fileobj(bucket, key, fp, ExtraArgs=extra_args)

                # go to start of file
                fp.seek(0)

                # Read and return all the contents
                return fp.read()

            except botocore.exceptions.ClientError:
                LOG.error(
                    "Unable to download Swagger document from S3 Bucket=%s Key=%s Version=%s", bucket, key, version
                )
                raise

    @staticmethod
    def _parse_s3_location(location):
        """
        Parses the given location input as a S3 Location and returns the file's bucket, key and version as separate
        values. Input can be in two different formats:

        1. Dictionary with ``Bucket``, ``Key``, ``Version`` keys
        2. String of S3 URI in format ``s3://<bucket>/<key>?versionId=<version>``

        If the input is not in either of the above formats, this method will return (None, None, None) tuple for all
        the values.

        Parameters
        ----------
        location : str or dict
            Location of the S3 file

        Returns
        -------
        str
            Name of the S3 Bucket. None, if bucket value was not found
        str
            Key of the file from S3. None, if key was not provided
        str
            Optional Version ID of the file. None, if version ID is not provided
        """

        bucket, key, version = None, None, None
        if isinstance(location, dict):
            # This is a S3 Location dictionary. Just grab the fields. It is very well possible that
            # this dictionary has none of the fields we expect. Return None if the fields don't exist.
            bucket, key, version = (location.get("Bucket"), location.get("Key"), location.get("Version"))

        elif isinstance(location, str) and location.startswith("s3://"):
            # This is a S3 URI. Parse it using a standard URI parser to extract the components

            parsed = urlparse(location)
            query = parse_qs(parsed.query)

            bucket = parsed.netloc
            key = parsed.path.lstrip("/")  # Leading '/' messes with S3 APIs. Remove it.

            # If there is a query string that has a single versionId field,
            # set the object version and return
            if query and "versionId" in query and len(query["versionId"]) == 1:
                version = query["versionId"][0]

        return bucket, key, version



"""
    Tests cookiecutter baking process and rendered content
"""


def test_project_tree(cookies):
    result = cookies.bake(extra_context={"project_name": "hello sam"})
    assert result.exit_code == 0
    assert result.exception is None
    assert result.project.basename == "hello sam"
    assert result.project.isdir()
    assert result.project.join("template.yaml").isfile()
    assert result.project.join("README.md").isfile()
    assert result.project.join("src").isdir()
    assert result.project.join("src", "main").isdir()
    assert result.project.join("src", "main", "java").isdir()
    assert result.project.join("src", "main", "java", "helloworld").isdir()
    assert result.project.join("src", "main", "java", "helloworld", "App.java").isfile()
    assert result.project.join("src", "main", "java", "helloworld", "GatewayResponse.java").isfile()
    assert result.project.join("src", "test", "java").isdir()
    assert result.project.join("src", "test", "java", "helloworld").isdir()
    assert result.project.join("src", "test", "java", "helloworld", "AppTest.java").isfile()


def test_app_content(cookies):
    result = cookies.bake(extra_context={"project_name": "my_lambda"})
    app_file = result.project.join("src", "main", "java", "helloworld", "App.java")
    app_content = app_file.readlines()
    app_content = "".join(app_content)

    contents = (
        "package helloword",
        "class App implements RequestHandler<Object, Object>",
        "https://checkip.amazonaws.com",
        "return new GatewayResponse",
        "getPageContents",
    )

    for content in contents:
        assert content in app_content



"""
Handles bundler properties as needed to modify the build process
"""
import logging
from copy import deepcopy
from typing import Dict, Optional

from samcli.lib.providers.provider import Stack
from samcli.lib.providers.sam_function_provider import SamFunctionProvider

LOG = logging.getLogger(__name__)

ESBUILD_PROPERTY = "esbuild"


class EsbuildBundlerManager:
    def __init__(self, stack: Stack, template: Optional[Dict] = None):
        self._stack = stack
        self._previous_template = template or dict()

    def esbuild_configured(self) -> bool:
        """
        Checks if esbuild is configured on any resource in a given stack
        :return: True if there is a function instance using esbuild as the build method
        """
        function_provider = SamFunctionProvider(
            [self._stack], use_raw_codeuri=True, ignore_code_extraction_warnings=True
        )
        functions = list(function_provider.get_all())
        for function in functions:
            if function.metadata and function.metadata.get("BuildMethod", "") == ESBUILD_PROPERTY:
                return True
        return False

    def set_sourcemap_metadata_from_env(self) -> Stack:
        """
        Checks if sourcemaps are set in lambda environment and updates build metadata accordingly.
        :return: Modified stack
        """
        modified_stack = deepcopy(self._stack)

        using_source_maps = False
        stack_resources = modified_stack.resources

        for name, resource in stack_resources.items():
            metadata = resource.get("Metadata", {})

            if not self._esbuild_in_metadata(metadata):
                continue

            node_option_set = self._is_node_option_set(resource)

            # check if Sourcemap is provided and append --enable-source-map if not set
            build_properties = metadata.get("BuildProperties", {})
            source_map = build_properties.get("Sourcemap", None)

            # check if --enable-source-map is provided and append Sourcemap: true if it is not set
            if source_map is None and node_option_set:
                LOG.info(
                    "\n--enable-source-maps set without Sourcemap, adding Sourcemap to"
                    " Metadata BuildProperties for %s",
                    name,
                )

                resource.setdefault("Metadata", {})
                resource["Metadata"].setdefault("BuildProperties", {})
                resource["Metadata"]["BuildProperties"]["Sourcemap"] = True

                using_source_maps = True

        if using_source_maps:
            self._warn_using_source_maps()

        return modified_stack

    def set_sourcemap_env_from_metadata(self) -> Dict:
        """
        Appends ``NODE_OPTIONS: --enable-source-maps``, if Sourcemap is set to true
        and sets Sourcemap to true if ``NODE_OPTIONS: --enable-source-maps`` is provided.
        :return: Dict containing deep-copied, updated template
        """
        using_source_maps = False
        invalid_node_option = False

        template = deepcopy(self._previous_template)
        template_resources = template.get("Resources", {})

        # We check the stack resources since they contain the global values, we modify the template
        stack_resources = self._stack.resources

        for name, stack_resource in stack_resources.items():
            metadata = stack_resource.get("Metadata", {})

            if not self._esbuild_in_metadata(metadata):
                continue

            node_option_set = self._is_node_option_set(stack_resource)

            template_resource = template_resources.get(name, {})

            # check if Sourcemap is provided and append --enable-source-map if not set
            build_properties = metadata.get("BuildProperties", {})
            source_map = build_properties.get("Sourcemap", None)

            if source_map and not node_option_set:
                LOG.info(
                    "\nSourcemap set without --enable-source-maps, adding"
                    " --enable-source-maps to function %s NODE_OPTIONS",
                    name,
                )

                template_resource.setdefault("Properties", {})
                template_resource["Properties"].setdefault("Environment", {})
                template_resource["Properties"]["Environment"].setdefault("Variables", {})
                existing_options = template_resource["Properties"]["Environment"]["Variables"].setdefault(
                    "NODE_OPTIONS", ""
                )

                # make sure the NODE_OPTIONS is a string
                if not isinstance(existing_options, str):
                    invalid_node_option = True
                else:
                    template_resource["Properties"]["Environment"]["Variables"]["NODE_OPTIONS"] = " ".join(
                        [existing_options, "--enable-source-maps"]
                    )

                using_source_maps = True

        if using_source_maps:
            self._warn_using_source_maps()

        if invalid_node_option:
            self._warn_invalid_node_options()

        return template

    @staticmethod
    def _esbuild_in_metadata(metadata: Dict) -> bool:
        """
        Checks if esbuild is configured in the function's metadata
        :param metadata: dict of metadata properties of a function
        :return: True if esbuild is configured, False otherwise
        """
        return bool(metadata.get("BuildMethod", "") == ESBUILD_PROPERTY)

    @staticmethod
    def _is_node_option_set(resource: Dict) -> bool:
        """
        Checks if the template has NODE_OPTIONS --enable-source-maps set

        Parameters
        ----------
        resource : Dict
            The resource dictionary to lookup if --enable-source-maps is set

        Returns
        -------
        bool
            True if --enable-source-maps is set, otherwise false
        """
        try:
            node_options = resource["Properties"]["Environment"]["Variables"]["NODE_OPTIONS"]

            return "--enable-source-maps" in node_options.split()
        except (KeyError, AttributeError):
            return False

    @staticmethod
    def _warn_invalid_node_options() -> None:
        """
        Log warning for invalid node options
        """
        LOG.info(
            "\nNODE_OPTIONS is not a string! As a result, the NODE_OPTIONS environment variable will "
            "not be set correctly, please make sure it is a string. "
            "Visit https://nodejs.org/api/cli.html#node_optionsoptions for more details.\n",
        )

    @staticmethod
    def _warn_using_source_maps() -> None:
        """
        Log warning telling user that node options will be set
        :return:
        """
        LOG.info(
            "\nYou are using source maps, note that this comes with a performance hit!"
            " Set Sourcemap to false and remove"
            " NODE_OPTIONS: --enable-source-maps to disable source maps.\n",
        )



"""Class container to hold common Service Responses"""

from flask import jsonify, make_response


class ServiceErrorResponses:

    _NO_LAMBDA_INTEGRATION = {"message": "No function defined for resource method"}
    _MISSING_AUTHENTICATION = {"message": "Missing Authentication Token"}
    _LAMBDA_FAILURE = {"message": "Internal server error"}

    HTTP_STATUS_CODE_501 = 501
    HTTP_STATUS_CODE_502 = 502
    HTTP_STATUS_CODE_403 = 403

    @staticmethod
    def lambda_failure_response(*args):
        """
        Helper function to create a Lambda Failure Response

        :return: A Flask Response
        """
        response_data = jsonify(ServiceErrorResponses._LAMBDA_FAILURE)
        return make_response(response_data, ServiceErrorResponses.HTTP_STATUS_CODE_502)

    @staticmethod
    def not_implemented_locally(message):
        """
        Constructs a Flask Response for for when a Lambda function functionality is
        not implemented

        :return: a Flask Response
        """
        exception_dict = {"message": message}
        response_data = jsonify(exception_dict)
        return make_response(response_data, ServiceErrorResponses.HTTP_STATUS_CODE_501)

    @staticmethod
    def lambda_not_found_response(*args):
        """
        Constructs a Flask Response for when a Lambda function is not found for an endpoint

        :return: a Flask Response
        """
        response_data = jsonify(ServiceErrorResponses._NO_LAMBDA_INTEGRATION)
        return make_response(response_data, ServiceErrorResponses.HTTP_STATUS_CODE_502)

    @staticmethod
    def route_not_found(*args):
        """
        Constructs a Flask Response for when a API Route (path+method) is not found. This is usually
        HTTP 404 but with API Gateway this is a HTTP 403 (https://forums.aws.amazon.com/thread.jspa?threadID=2166840)

        :return: a Flask Response
        """
        response_data = jsonify(ServiceErrorResponses._MISSING_AUTHENTICATION)
        return make_response(response_data, ServiceErrorResponses.HTTP_STATUS_CODE_403)



""" configure Schemas client based on AWS configuration provided by user """

import click
from boto3.session import Session

from samcli.commands.local.cli_common.user_exceptions import ResourceNotFound


def get_aws_configuration_choice():
    """
    Allow user to select their AWS Connection configuration (profile/region)
    :return: AWS profile and region dictionary
    """
    session = Session()
    profile = session.profile_name
    region = session.region_name

    message = f"\nDo you want to use the default AWS profile [{profile}] and region [{region}]?"
    # Avoid prompting if there isn't a region, the user must supply the information
    needs_edit = (not region) or (not click.confirm(message, default=True))

    schemas_available_regions_name = session.get_available_regions("schemas")

    if needs_edit:
        available_profiles = session.available_profiles
        profile = _get_aws_profile_choice(available_profiles)
        # Reinitialize the session with the new profile to get the new region
        session = Session(profile_name=profile)
        region = _get_aws_region_choice(schemas_available_regions_name, session.region_name)
    else:
        # session.profile_name will return 'default' if no profile is found,
        # but botocore itself will fail if you pass it in, when one is not configured
        profile = None

    return {"profile": profile, "region": region}


def _get_aws_profile_choice(available_profiles):
    if not available_profiles:
        raise ResourceNotFound("No configured AWS profile found.")

    # Convert list of available profiles (strings) into a list of click.Choice
    # index/string value tuples
    profile_choices = list(map(str, range(1, len(available_profiles) + 1)))
    profile_choice_num = 1

    click.echo("\nWhich AWS profile do you want to use?")

    for profile in available_profiles:
        msg = str(profile_choice_num) + " - " + profile
        click.echo("\t" + msg)
        profile_choice_num = profile_choice_num + 1

    profile_choice = click.prompt("Profile", type=click.Choice(profile_choices), show_choices=False)
    return available_profiles[int(profile_choice) - 1]


def _get_aws_region_choice(available_regions_name, region):
    if not available_regions_name:
        raise ResourceNotFound(
            "No AWS region found for AWS schemas service. This should not be possible, please raise an issue."
        )
    cli_display_regions = dict()

    for available_region_name in available_regions_name:
        region_prefix = available_region_name.rsplit("-", 1)[0]
        if region_prefix not in cli_display_regions:
            cli_display_regions[region_prefix] = available_region_name
        else:
            cli_display_regions[region_prefix] = cli_display_regions.get(region_prefix) + "," + available_region_name

    click.echo("\nWhich region do you want to use for your schema registry?")
    click.echo("# Partial list of AWS regions")
    click.echo("#")

    for msg in cli_display_regions.values():
        click.echo("# " + msg)

    region_choice = click.prompt(f"Region [{region}]", type=str, show_choices=False)
    return region_choice


def get_schemas_client(profile, region):
    if profile:
        session = Session(profile_name=profile)
    else:
        session = Session()
    return session.client("schemas", region_name=region)



"""
CLI command for "traces" command
"""
import logging

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, pass_context, print_cmdline_args
from samcli.cli.main import common_options as cli_framework_options
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands._utils.options import common_observability_options
from samcli.lib.observability.util import OutputOption
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version

LOG = logging.getLogger(__name__)

HELP_TEXT = """
Use this command to fetch AWS X-Ray traces generated by your stack.\n
\b
Run the following command to fetch X-Ray traces by ID.
$ sam traces --trace-id tracing-id-1 --trace-id tracing-id-2
\b
Run the following command to tail X-Ray traces as they become available.
$ sam traces --tail
"""


@click.command("traces", help=HELP_TEXT, short_help="Fetch AWS X-Ray traces")
@configuration_option(provider=TomlProvider(section="parameters"))
@click.option(
    "--trace-id",
    "-ti",
    multiple=True,
    help="Fetch specific trace by providing its id",
)
@common_observability_options
@cli_framework_options
@aws_creds_options
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
@command_exception_handler
def cli(
    ctx,
    trace_id,
    start_time,
    end_time,
    tail,
    output,
    config_file,
    config_env,
):
    """
    `sam traces` command entry point
    """
    do_cli(trace_id, start_time, end_time, tail, output, ctx.region)


def do_cli(trace_ids, start_time, end_time, tailing, output, region):
    """
    Implementation of the ``cli`` method
    """
    from datetime import datetime

    import boto3

    from samcli.commands.logs.logs_context import parse_time
    from samcli.commands.traces.traces_puller_factory import generate_trace_puller
    from samcli.lib.utils.boto_utils import get_boto_config_with_user_agent

    sanitized_start_time = parse_time(start_time, "start-time")
    sanitized_end_time = parse_time(end_time, "end-time") or datetime.utcnow()

    boto_config = get_boto_config_with_user_agent(region_name=region)
    xray_client = boto3.client("xray", config=boto_config)

    # generate puller depending on the parameters
    puller = generate_trace_puller(xray_client, OutputOption(output) if output else OutputOption.text)

    if trace_ids:
        puller.load_events(trace_ids)
    elif tailing:
        puller.tail(sanitized_start_time)
    else:
        puller.load_time_period(sanitized_start_time, sanitized_end_time)



"""
Class that provides layers from a given SAM template
"""
import logging
from typing import Dict, List, Optional

from samcli.lib.utils.resources import AWS_LAMBDA_LAYERVERSION, AWS_SERVERLESS_LAYERVERSION

from .provider import LayerVersion, Stack
from .sam_base_provider import SamBaseProvider
from .sam_stack_provider import SamLocalStackProvider

LOG = logging.getLogger(__name__)


class SamLayerProvider(SamBaseProvider):
    """
    Fetches and returns Layers from a SAM Template. The SAM template passed to this provider is assumed to be valid,
    normalized and a dictionary.

    It may or may not contain a layer.
    """

    def __init__(self, stacks: List[Stack], use_raw_codeuri: bool = False) -> None:
        """
        Initialize the class with SAM template data. The SAM template passed to this provider is assumed
        to be valid, normalized and a dictionary. It should be normalized by running all pre-processing
        before passing to this class. The process of normalization will remove structures like ``Globals``, resolve
        intrinsic functions etc.
        This class does not perform any syntactic validation of the template.

        After the class is initialized, any changes to the ``template_dict`` will not be reflected in here.
        You need to explicitly update the class with new template, if necessary.

        Parameters
        ----------
        :param dict stacks: List of stacks layers are extracted from
        :param bool use_raw_codeuri: Do not resolve adjust core_uri based on the template path, use the raw uri.
            Note(xinhol): use_raw_codeuri is temporary to fix a bug, and will be removed for a permanent solution.
        """
        self._stacks = stacks
        self._use_raw_codeuri = use_raw_codeuri

        self._layers = self._extract_layers()

    def get(self, name: str) -> Optional[LayerVersion]:
        """
        Returns the layer with given name or logical id.
        If it is in a nested stack, name can be prefixed with stack path to avoid ambiguity

        Parameters
        ----------
        name: name or logical id of the layer.

        Returns
        -------
        LayerVersion object of one layer.

        """
        if not name:
            raise ValueError("Layer name is required")

        for layer in self._layers:
            if name in (layer.full_path, layer.layer_id, layer.name):
                return layer
        return None

    def get_all(self) -> List[LayerVersion]:
        """
        Returns all Layers in template
        Returns
        -------
        [LayerVersion] list of layer version objects.
        """
        return self._layers

    def _extract_layers(self) -> List[LayerVersion]:
        """
        Extracts all resources with Type AWS::Lambda::LayerVersion and AWS::Serverless::LayerVersion and return a list
        of those resources.
        """
        layers = []
        for stack in self._stacks:
            for name, resource in stack.resources.items():
                # In the list of layers that is defined within a template, you can reference a LayerVersion resource.
                # When running locally, we need to follow that Ref so we can extract the local path to the layer code.
                resource_type = resource.get("Type")
                resource_properties = resource.get("Properties", {})

                if resource_type in [AWS_LAMBDA_LAYERVERSION, AWS_SERVERLESS_LAYERVERSION]:
                    code_property_key = SamBaseProvider.CODE_PROPERTY_KEYS[resource_type]
                    if SamBaseProvider._is_s3_location(resource_properties.get(code_property_key)):
                        # Content can be a dictionary of S3 Bucket/Key or a S3 URI, neither of which are supported
                        SamBaseProvider._warn_code_extraction(resource_type, name, code_property_key)
                        continue
                    codeuri = SamBaseProvider._extract_codeuri(resource_properties, code_property_key)

                    compatible_runtimes = resource_properties.get("CompatibleRuntimes")
                    compatible_architectures = resource_properties.get("CompatibleArchitectures", None)

                    metadata = resource.get("Metadata", None)
                    layers.append(
                        self._convert_lambda_layer_resource(
                            stack, name, codeuri, compatible_runtimes, metadata, compatible_architectures
                        )
                    )
        return layers

    def _convert_lambda_layer_resource(
        self,
        stack: Stack,
        layer_logical_id: str,
        codeuri: str,
        compatible_runtimes: Optional[List[str]],
        metadata: Optional[Dict],
        compatible_architectures: Optional[List[str]],
    ) -> LayerVersion:
        """
        Convert layer resource into {LayerVersion} object.
        Parameters
        ----------
        stack
        layer_logical_id
            LogicalID of resource.
        codeuri
            codeuri of the layer
        compatible_runtimes
            list of compatible runtimes
        metadata
            dictionary of layer metadata
        compatible_architectures
            list of compatible architecture
        Returns
        -------
        LayerVersion
            The layer object
        """
        if codeuri and not self._use_raw_codeuri:
            LOG.debug("--base-dir is not presented, adjusting uri %s relative to %s", codeuri, stack.location)
            codeuri = SamLocalStackProvider.normalize_resource_path(stack.location, codeuri)

        return LayerVersion(
            layer_logical_id,
            codeuri,
            compatible_runtimes,
            metadata,
            compatible_architectures=compatible_architectures,
            stack_path=stack.stack_path,
        )



"""
Retry decorator to retry decorated function based on Exception with exponential backoff and number of attempts built-in.
"""
import math
import time
from functools import wraps


def retry(exc, attempts=3, delay=0.05, exc_raise=Exception, exc_raise_msg=""):
    """
    Retry decorator which defaults to 3 attempts based on exponential backoff
    and a delay of 50ms.
    After retries are exhausted, a custom Exception and Error message are raised.

    :param exc: Exception to be caught for retry
    :param attempts: number of attempts before exception is allowed to be raised.
    :param delay: an initial delay which will exponentially increase based on the retry attempt.
    :param exc_raise: Final Exception to raise.
    :param exc_raise_msg: Final message for the Exception to be raised.
    :return:
    """

    def retry_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            remaining_attempts = attempts
            retry_attempt = 1
            while remaining_attempts >= 1:
                try:
                    return func(*args, **kwargs)
                except exc:
                    time.sleep(math.pow(2, retry_attempt) * delay)
                    retry_attempt = retry_attempt + 1
                    remaining_attempts = remaining_attempts - 1
            raise exc_raise(exc_raise_msg)

        return wrapper

    return retry_wrapper
